Package: rapidcodeR
Title: Optimized Data Analysis System for AI-Based Text Processing
Version: 0.1.1
Authors@R: c(
    person("Gabriel", "Lönn", , "g.e.lonn@stv.uio.no", role = c("aut", "cre")),
    person("Sebastian", "Schutte", , role = "ctb", comment = "Original code and package idea provided by.")
    )
Description: Extracts machine-readable variables from natural language text using
    AI APIs. Optimized for speed and cost efficiency through parallel processing
    and direct CSV-formatted responses from language models. Supports multiple AI
    providers with robust error handling and automatic retry mechanisms for
    failed extractions.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: stringr, openai, groqR, dplyr, rlang, parallel, future,
        future.apply
Suggests: testthat (>= 3.0.0), irr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 12:14:21 UTC; gelon
Author: Gabriel Lönn [aut, cre],
  Sebastian Schutte [ctb] (Original code and package idea provided by.)
Maintainer: Gabriel Lönn <g.e.lonn@stv.uio.no>
Repository: CRAN
Date/Publication: 2026-05-08 17:03:24 UTC
Built: R 4.5.2; ; 2026-05-08 20:21:19 UTC; unix

Package: rapidcodeR
Title: Optimized Data Analysis System for AI-Based Text Processing
Version: 0.1.1
Authors@R: c(
    person("Gabriel", "Lönn", , "g.e.lonn@stv.uio.no", role = c("aut", "cre")),
    person("Sebastian", "Schutte", , role = "ctb", comment = "Original code and package idea provided by.")
    )
Description: Extracts machine-readable variables from natural language text using
    AI APIs. Optimized for speed and cost efficiency through parallel processing
    and direct CSV-formatted responses from language models. Supports multiple AI
    providers with robust error handling and automatic retry mechanisms for
    failed extractions.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: stringr, openai, groqR, dplyr, rlang, parallel, future,
        future.apply
Suggests: testthat (>= 3.0.0), irr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 12:14:21 UTC; gelon
Author: Gabriel Lönn [aut, cre],
  Sebastian Schutte [ctb] (Original code and package idea provided by.)
Maintainer: Gabriel Lönn <g.e.lonn@stv.uio.no>
Repository: CRAN
Date/Publication: 2026-05-08 17:03:24 UTC
Built: R 4.5.2; ; 2026-05-08 19:07:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(ask_groq)
export(ask_openai)
export(calculate_overlap)
export(main_func)
export(make_value_row)
export(parallel_execute)
export(set_api_specs)
export(set_coding_instruction)
export(set_parameters)
import(future)
import(future.apply)
importFrom(dplyr,"%>%")
importFrom(dplyr,.data)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,slice_sample)
importFrom(rlang,sym)
importFrom(stats,complete.cases)
importFrom(stats,setNames)
importFrom(stringr,str_remove)

#' Main Text Processing Function for AI-Based Variable Extraction
#'
#' This function processes batches of text data using AI models to extract
#' machine-readable variables. It implements robust error handling and retry
#' logic to ensure reliable processing even with API failures.
#'
#' @importFrom dplyr filter slice_sample bind_rows
#' @importFrom dplyr "%>%"
#' @importFrom rlang sym
#'
#' @param df Data frame. Input data subset containing text to be processed.
#' @param to_code_max_id Integer. Maximum number of posts to process in this batch.
#' @param n_post Integer. Number of posts to process per API call (typically 15).
#' @param provider Character. AI provider to use, either "OpenAI" or "Groq".
#' @param worker_env Environment or NULL. When running in a parallel worker, the worker's
#'   package environment. If NULL (default), \code{.package_env} is used.
#'
#' @details
#' The function implements a robust processing loop that:
#' \itemize{
#'   \item Samples posts randomly to avoid processing order bias
#'   \item Makes API calls in manageable batches
#'   \item Validates and cleans AI responses
#'   \item Handles API failures with counter and early stopping
#'   \item Tracks missing IDs for potential reprocessing
#'   \item Collects dataframes into a list for later binding
#' }
#'
#' The AI models are instructed to return data in CSV format for efficient parsing.
#' The function expects responses with exactly N columns as set by set_parameters().
#'
#' @return List of data frames. Each element contains successfully processed results
#'   with extracted variables. Returns empty list if no successful processing.
#'
#' @examples
#' \dontrun{
#' # For set_parameters(n_variables = 6):
#' instruction <- "Extract variables: var1,var2,var3,var4,var5,var6"
#' result <- main_func(
#'   df = my_data_subset,
#'   to_code_max_id = 30,
#'   n_post = 15,
#'   provider = "OpenAI"
#' )
#' processed_data <- bind_rows(result)  # Combine all dataframes
#' }
#'
#' @seealso [gpt_func()], [groq_func()], [make_value_row()]
#' @export
main_func <- function(df, to_code_max_id, n_post, provider, worker_env = NULL) {

  env <- if (is.null(worker_env)) .package_env else worker_env

  # Input validation
  if (!is.data.frame(df)) {
    stop("df must be a data frame")
  }

  # Get column parameters from package environment
  id_column <- get("id_column", envir = env)
  text_column <- get("text_column", envir = env)

  # Validate that the specified columns exist
  if (id_column > ncol(df)) {
    stop("id_column (", id_column, ") is greater than the number of columns in df (", ncol(df), ")")
  }

  if (text_column > ncol(df)) {
    stop("text_column (", text_column, ") is greater than the number of columns in df (", ncol(df), ")")
  }

  # Get column names for easier reference
  # ID column is always column 1 (internal_id)
  id_col_name <- names(df)[1]
  # User's ID column is adjusted by +1 because internal_id column was added
  internal_id_column <- id_column + 1
  user_id_col_name <- names(df)[internal_id_column]
  # Text column is adjusted by +1 because internal_id column was added
  internal_text_column <- text_column + 1
  text_col_name <- names(df)[internal_text_column]

  if (!is.numeric(to_code_max_id) || to_code_max_id <= 0) {
    stop("to_code_max_id must be a positive number")
  }

  if (!is.numeric(n_post) || n_post <= 0) {
    stop("n_post must be a positive number")
  }

  if (!provider %in% c("OpenAI", "Groq")) {
    stop("provider must be either 'OpenAI' or 'Groq'")
  }

  # Get multi_response from package environment
  multi_response <- if (exists("multi_response", envir = env)) {
    get("multi_response", envir = env)
  } else {
    FALSE  # Default fallback
  }

  if (!is.logical(multi_response) || length(multi_response) != 1) {
    stop("multi_response must be a single logical value (TRUE or FALSE)")
  }

  # Validate multi_response constraint
  if (multi_response && n_post != 1) {
    stop("multi_response = TRUE can only be used with n_post = 1")
  }

  coding_instruction <- get("coding_instruction", envir = env)

  n_variables <- as.numeric(get("n_variables", envir = env))

  `%!in%` <- Negate(`%in%`)

  coded_ids <- as.vector(NULL)

  # Initialize list to collect dataframes
  result_list <- list()


  # Single while loop):
  while(to_code_max_id > 0) {

    # Get the current batch of data to process using dynamic column names
    data_sample <- df %>%
      filter(!!sym(id_col_name) %!in% coded_ids) %>%
      slice_sample(n = min(n_post, to_code_max_id), replace = FALSE)

      if (nrow(data_sample) == 0) {
        break
      }

      # Store the internal IDs for this batch
      batch_internal_ids <- data_sample[[id_col_name]]

      # Store user IDs for this batch
      batch_user_ids <- data_sample[[user_id_col_name]]

      # Remove internal_id column for API processing (keep user_id and other columns)
      data_for_api <- data_sample[, -1, drop = FALSE]

      # Process with API
      if (provider == "OpenAI") {
        result_df <- gpt_func(data_for_api, n_post, coding_instruction, worker_env = worker_env)
      } else if (provider == "Groq") {
        result_df <- groq_func(data_for_api, n_post, coding_instruction, worker_env = worker_env)
      }


      # Process API response and add internal_id back
      if (!is.null(result_df) && nrow(result_df) > 0) {

        # Handle row count validation based on multi_response setting
        if (!multi_response && nrow(result_df) != length(batch_internal_ids)) {
          # Skip this iteration if multi_response=FALSE but row counts don't match
          # Don't mark IDs as processed since this batch failed
          to_code_max_id <- to_code_max_id - nrow(data_sample)
          next
        }

        # Add IDs back to the results
        if (multi_response) {
          # For multi_response=TRUE, all rows get the same user_id and internal_id
          result_df$user_id <- batch_user_ids[1]
          result_df$internal_id <- batch_internal_ids[1]
        } else {
          # For multi_response=FALSE, each row gets its corresponding user_id and internal_id
          # Only assign IDs for the rows that actually exist in the result
          n_result_rows <- nrow(result_df)
          #result_df$user_id <- batch_user_ids[1:n_result_rows]
          result_df$internal_id <- batch_internal_ids[1:n_result_rows]
        }

        # Move internal_id to first column
        result_df <- result_df[, c("internal_id", setdiff(names(result_df), c("internal_id")))]

        # Add flag for multiple responses (1 if multiple rows, 0 otherwise)
        result_df$multiple_response_flag <- ifelse(nrow(result_df) > 1, 1, 0)

        # Add to result list
        result_list[[length(result_list) + 1]] <- result_df

        # Update tracking variables - only mark IDs as processed if we successfully got results
        coded_ids <- c(coded_ids, data_sample[[id_col_name]])
        to_code_max_id <- to_code_max_id - nrow(data_sample)
      } else {
        # No results from API, skip to next batch
        # Don't mark IDs as processed since this batch failed
        to_code_max_id <- to_code_max_id - nrow(data_sample)
      }
  }

  return(result_list)
}

#' rapidcodeR: Optimized Data Analysis System for AI-Based Text Processing
#'
#' A high-performance R package for extracting machine-readable variables from
#' natural language text using AI APIs. The package is optimized for speed and
#' cost efficiency through parallel processing and direct CSV-formatted responses
#' from language models.
#'
#' @section Funding:
#' This work is financed in large part by the Research Council of Norway
#' (grant #324931).
#'
#' @section Key Features:
#' \itemize{
#'   \item \strong{Parallel Processing}: Distributes work across multiple CPU cores for maximum speed
#'   \item \strong{Cost Optimization}: Uses batch processing to minimize API calls
#'   \item \strong{Multi-Provider Support}: Works with both OpenAI GPT and GROQ AI services
#'   \item \strong{Robust Error Handling}: Automatic retry mechanisms for failed extractions
#'   \item \strong{CSV-Optimized Output}: AI models return structured data for efficient parsing
#'   \item \strong{Missing Data Recovery}: Secondary processing stage for failed extractions
#' }
#'
#' @section Main Functions:
#' \describe{
#'   \item{\code{\link{parallel_execute}}}{Main function for parallel text processing}
#'   \item{\code{\link{set_api_specs}}}{Configure API specifications for AI services}
#'   \item{\code{\link{set_coding_instruction}}}{Define how AI should extract variables}
#'   \item{\code{\link{set_parameters}}}{Configure processing parameters}
#'   \item{\code{\link{calculate_overlap}}}{Assess reliability across multiple runs}
#' }
#'
#' @section Workflow:
#' \enumerate{
#'   \item Set up API specifications using \code{set_api_specs()}
#'   \item Configure processing parameters with \code{set_parameters()} (optional)
#'   \item Define extraction instructions with \code{set_coding_instruction()}
#'   \item Process data with \code{parallel_execute()}
#'   \item Optionally assess reliability with \code{calculate_overlap()}
#' }
#'
#' @section Data Format:
#' Input data should contain columns specified by \code{set_parameters()}:
#' \itemize{
#'   \item ID column: Unique identifier for each text (default column 1)
#'   \item Text column: The text content to be processed (default column 2)
#' }
#'
#' Output contains configurable columns: id, language, text, and N extracted variables (Var1-VarN).
#'
#' @examples
#' \dontrun{
#' # Basic workflow
#' library(rapidcodeR)
#'
#' # 1. Set up API specifications
#' set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_openai_key")
#'
#' # 2. Configure processing parameters (optional)
#' set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ",")
#'
#' # 3. Define what to extract
#' instruction <- paste(
#'   "Extract sentiment and topic from each post.",
#'   "Return as CSV: id,language,text,sentiment,confidence,topic,relevance"
#' )
#' set_coding_instruction(instruction)
#'
#' # 4. Process your data
#' results <- parallel_execute(
#'   test_data = my_text_data,
#'   slicing_n = 240,     # Process 240 random posts
#'   cores = 4,           # Use 4 CPU cores
#'   # Provider determined by set_api_specs()
#' )
#'
#' # 5. Check reliability (optional)
#' run1 <- parallel_execute(my_data, 100, cores = 4, seed = 123)
#' run2 <- parallel_execute(my_data, 100, cores = 4, seed = 123)
#' agreement <- calculate_overlap(list(run1, run2))
#' }
#'
#' @keywords internal
"_PACKAGE"

# Package imports are handled by DESCRIPTION Imports field
# These library calls are only needed during development

.package_env <- new.env(parent = emptyenv())

# Package imports are handled by DESCRIPTION Imports field


#' Set API Specifications for AI Models
#'
#' This function sets the API specifications including provider, model, temperature,
#' and API key for AI model interactions. It replaces the need for separate
#' set_api_keys calls and allows for more flexible configuration.
#'
#' @param provider Character. The AI provider to use, either "OpenAI" or "Groq".
#' @param model Character. The specific model to use (e.g., "gpt-4", "gpt-3.5-turbo", "llama-3.3-70b-versatile").
#' @param temp Numeric. Temperature parameter controlling randomness (0-1). Lower values make output more deterministic.
#' @param api_key Character. The API key for the specified provider.
#' @param multi_response Logical. If TRUE, allows AI to return multiple rows per input (only works with n_post=1).
#'   If FALSE (default), AI must return exactly one row per input.
#'
#' @details
#' This function stores the API specifications in the package's internal environment
#' and sets the appropriate environment variables for API authentication. The
#' specifications are automatically retrieved by processing functions.
#'
#' @return Character. A confirmation message.
#'
#' @examples
#' \dontrun{
#' # Set OpenAI specifications
#' set_api_specs(
#'   provider = "OpenAI",
#'   model = "gpt-4",
#'   temp = 0.7,
#'   api_key = "your_openai_key"
#' )
#'
#' # Set GROQ specifications
#' set_api_specs(
#'   provider = "Groq",
#'   model = "llama-3.3-70b-versatile",
#'   temp = 0.3,
#'   api_key = "your_groq_key"
#' )
#' }
#'
#' @seealso [set_coding_instruction()], [set_parameters()]
#' @export
set_api_specs <- function(provider, model, temp, api_key, multi_response = FALSE) {

  # Input validation
  if (!is.character(provider) || length(provider) != 1) {
    stop("provider must be a single character string")
  }

  if (!provider %in% c("OpenAI", "Groq")) {
    stop("provider must be either 'OpenAI' or 'Groq'")
  }

  if (!is.character(model) || length(model) != 1) {
    stop("model must be a single character string")
  }

  if (!is.numeric(temp) || length(temp) != 1 || temp < 0 || temp > 1) {
    stop("temp must be a single numeric value between 0 and 1")
  }

  if (!is.character(api_key) || length(api_key) != 1) {
    stop("api_key must be a single character string")
  }

  if (!is.logical(multi_response) || length(multi_response) != 1) {
    stop("multi_response must be a single logical value (TRUE or FALSE)")
  }

  # Set environment variables based on provider
  if (provider == "OpenAI") {
    Sys.setenv(OPENAI_API_KEY = api_key)
  } else if (provider == "Groq") {
    assign("GROQ_API_KEY", api_key, envir = .package_env)
  }

  # Store specifications in package environment
  assign("api_provider", provider, envir = .package_env)
  assign("api_model", model, envir = .package_env)
  assign("api_temp", temp, envir = .package_env)
  assign("multi_response", multi_response, envir = .package_env)

  out <- paste("API specs set: provider =", provider,
               ", model =", model,
               ", temperature =", temp,
               ", multi_response =", multi_response,
               ", API key set")

  return(out)
}

#' Set Coding Instructions for AI Models
#'
#' This function sets the coding instructions that will be used by AI models
#' to extract machine-readable variables from natural language text. The instruction
#' defines how the AI should format and structure its responses.
#'
#' @param instruction Character. A string containing the coding instruction that
#'   tells the AI model how to process and format the text data. Should specify
#'   the expected output format (typically CSV-structured responses).
#'
#' @details
#' The coding instruction is stored in the package's internal environment and
#' is automatically retrieved by processing functions. This instruction typically
#' defines:
#' \itemize{
#'   \item Expected output format (e.g., CSV structure)
#'   \item Variable definitions and coding schemes
#'   \item Response formatting requirements
#' }
#'
#' @return Character. A confirmation message indicating the instruction has been set.
#'
#' @examples
#' \dontrun{
#' instruction <- "Extract sentiment and topics from posts. Return as CSV: id,sentiment,topic"
#' set_coding_instruction(instruction)
#' }
#'
#' @export
set_coding_instruction <- function(instruction) {
  assign("coding_instruction", instruction, envir = .package_env)

  out <- "Coding instruction has been set!"

  return(out)
}

#' Set Processing Parameters for AI Models
#'
#' This function configures the parameters that will be used by AI models
#' to extract machine-readable variables from natural language text.
#'
#' @param n_variables Integer. Number of columns to extract (default is 9).
#'   This determines the total number of columns in the final dataset.
#'   Must be between 1 and 20.
#' @param id_column Integer. The column number containing the unique ID for each
#'   observation (default is 1). Must be a positive integer.
#' @param text_column Integer. The column number containing the text content
#'   to be processed (default is 2). Must be a positive integer.
#' @param sep Character. Separator used in API response parsing (default is ";").
#'   Must be one of: ";" (semicolon), "," (comma), or "|" (vertical bar).
#'
#' @details
#' The function sets the expected parameters in the package's internal
#' environment. This affects:
#' \itemize{
#'   \item Response parsing and validation
#'   \item Column naming (Var1, Var2, ..., VarN)
#'   \item Data structure expectations across all processing functions
#'   \item Column identification for text content and unique IDs
#'   \item Separator used for parsing API responses
#' }
#'
#' The package automatically adds a unique internal ID column to each row during processing.
#' This internal ID is used for tracking and is removed from the final output.
#'
#' The total expected columns will be n_variables:
#' \itemize{
#'   \item All columns: Var1, Var2, ..., VarN
#' }
#'
#' @return Character. A confirmation message indicating the parameters have been set.
#'
#' @examples
#' \dontrun{
#' # Extract 4 columns, ID in column 1, text in column 2, semicolon separator
#' set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
#'
#' # Extract 10 columns, ID in column 3, text in column 5, comma separator
#' set_parameters(n_variables = 10, id_column = 3, text_column = 5, sep = ",")
#'
#' # Extract 6 columns, ID in column 2, text in column 3, vertical bar separator
#' set_parameters(n_variables = 6, id_column = 2, text_column = 3, sep = "|")
#' }
#'
#' @export
set_parameters <- function(n_variables = 9, id_column = 1, text_column = 2, sep = ";") {
  # Input validation
  if (!is.numeric(n_variables) || n_variables < 1 || n_variables > 20) {
    stop("n_variables must be a positive number between 1 and 20")
  }

  if (!is.numeric(id_column) || id_column <= 0) {
    stop("id_column must be a positive number")
  }

  if (!is.numeric(text_column) || text_column <= 0) {
    stop("text_column must be a positive number")
  }

  # Validate separator
  valid_separators <- c(";", ",", "|")
  if (!sep %in% valid_separators) {
    stop("sep must be one of: ';' (semicolon), ',' (comma), or '|' (vertical bar)")
  }

  # Store parameters in package environment
  assign("n_variables", as.integer(n_variables), envir = .package_env)
  assign("id_column", as.integer(id_column), envir = .package_env)
  assign("text_column", as.integer(text_column), envir = .package_env)
  assign("sep", sep, envir = .package_env)

  out <- paste("Parameters set: n_variables =", n_variables,
               ", id_column =", id_column,
               ", text_column =", text_column,
               ", sep = '", sep, "'",
               " (internal ID column will be added automatically)")

  return(out)
}

test_that("Flexibility tests for rapidcodeR", {

  # Skip if API keys are not set
  skip_if_not(is_api_key_available(), "OpenAI API key not set")

  # Test API connectivity first
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  test_result <- tryCatch({
    ask_openai("What is the capital of Norway?")
  }, error = function(e) {
    return(NA)
  })

  # Skip test if API connection fails
  skip_if(is.na(test_result), "OpenAI API test failed - check API key and internet connection")

  # Create diverse test data
  test_data <- data.frame(
    post_id = 1:35,
    full_post = c(
      # Short posts
      "Great day!", "Not good.", "Excellent service.",

      # Medium posts
      "This product is amazing and I highly recommend it to everyone.",
      "I had a terrible experience with customer service yesterday.",
      "The weather is beautiful today, perfect for a walk.",

      # Long posts
      "I've been using this software for months now and I must say it has completely transformed my workflow. The interface is intuitive, the features are comprehensive, and the performance is outstanding. I would definitely recommend it to anyone looking for a reliable solution.",
      "Yesterday I went to the new restaurant downtown and was disappointed with both the food quality and service. The wait time was excessive, the staff seemed overwhelmed, and the food was cold when it arrived. I don't think I'll be returning.",
      "The conference was incredibly informative with speakers from various industries sharing their insights on technology trends, market analysis, and future predictions. I learned so much about emerging technologies and their potential applications.",

      # Multilingual content (simplified for testing)
      "Bonjour, comment allez-vous? Je suis très content.",
      "Hola, ¿cómo estás? Todo está muy bien.",
      "Guten Tag, wie geht es Ihnen? Alles ist gut.",

      # Posts with special characters
      "Check out this amazing deal: 50% off! 🎉 #sale #discount",
      "Meeting @ 3:00 PM tomorrow. Please bring your laptops & notebooks.",
      "The results are in: 95% success rate! (That's amazing!)",

      # Technical content
      "The API endpoint /api/v1/users returns a JSON response with user data including id, name, email, and timestamp fields.",
      "Error 404: Resource not found. Please check the URL and try again.",
      "Database query executed successfully. 1,247 records updated in 0.3 seconds.",

      # Emotional content
      "I'm so excited about this opportunity! Can't wait to get started.",
      "Feeling really frustrated with this situation. Nothing seems to be working.",
      "This makes me so happy! Thank you for making my day better.",

      # Question/answer format
      "What is the best way to optimize database performance?",
      "How do I configure the security settings for maximum protection?",
      "Where can I find documentation for the new API features?",

      # Mixed content
      "Product review: ⭐⭐⭐⭐⭐ Excellent quality, fast shipping, great customer service. Highly recommend!",
      "Technical issue resolved: Updated firmware to v2.1.3, rebooted system, cleared cache. All systems operational.",
      "Event update: Conference moved to Room 204B. Lunch will be served at 12:30 PM. See you there!",

      # Edge cases
      "a",  # Very short
      "12345",  # Numbers only
      "!@#$%^&*()",  # Special characters only
      "",  # Empty (will be filtered out)
      "   ",  # Whitespace only (will be filtered out)
      "This is a very long post that contains multiple sentences and should test the system's ability to handle longer text content. It includes various punctuation marks, numbers like 123 and 456, and special characters like @, #, and $. The post continues with more content to ensure it's long enough to be processed properly by the AI model.",
      "Mixed content: English text with 中文 characters and العربية script.",
      "Email: user@example.com, Phone: +1-555-123-4567, Website: https://example.com"
    )
  )

  # Test 1: Different variable counts
  cat("\nTesting different variable counts...\n")

  set_api_specs(provider = "OpenAI", model = "gpt-5", temp = 1,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  # Test with 3 variables
  set_parameters(n_variables = 3, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract exactly 3 variables from each text: sentiment;confidence;topic. Return ONLY the values separated by semicolons, no explanations or labels. Example: positive;0.8;technology")

  result_3vars <- parallel_execute(
    test_data = test_data[1:10, ],
    slicing_n = 10,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(ncol(result_3vars) == 3, "Should extract 3 variables")
  expect_true(all(c("Var1", "Var2", "Var3") %in% names(result_3vars)),
              "Should have correct column names for 3 variables")

  # Test with 6 variables
  set_parameters(n_variables = 6, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract exactly 6 variables from each text: sentiment;confidence;topic;emotion;intensity;relevance. Return ONLY the values separated by semicolons, no explanations or labels. Example: positive;0.8;technology;joy;high;high")

  result_6vars <- parallel_execute(
    test_data = test_data[1:10, ],
    slicing_n = 10,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(ncol(result_6vars) == 6, "Should extract 6 variables")
  expect_true(all(c("Var1", "Var2", "Var3", "Var4", "Var5", "Var6") %in% names(result_6vars)),
              "Should have correct column names for 6 variables")


  # Test 2: Different separators
  cat("\nTesting different separators...\n")

  # Test with comma separator
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ",")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  result_comma <- parallel_execute(
    test_data = test_data[1:8, ],
    slicing_n = 8,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_comma), "Comma separator should work")

  # Test with pipe separator
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = "|")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  result_pipe <- parallel_execute(
    test_data = test_data[1:8, ],
    slicing_n = 8,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_pipe), "Pipe separator should work")

  # Test 3: Different core counts
  cat("\nTesting different core counts...\n")

  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  # Test with 1 core
  result_1core <- parallel_execute(
    test_data = test_data[1:8, ],
    slicing_n = 8,
    cores = 1,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_1core), "Single core should work")

  # Test with 2 cores (if available)
  if (parallel::detectCores() >= 2) {
    result_2cores <- parallel_execute(
      test_data = test_data[1:8, ],
      slicing_n = 8,
      cores = 2,
      multi_core = FALSE
    )

    expect_true(is.data.frame(result_2cores), "Two cores should work")
  }

  # Test 4: Different text content types
  cat("\nTesting different text content types...\n")

  # Test with short posts
  short_posts <- test_data[1:5, ]
  result_short <- parallel_execute(
    test_data = short_posts,
    slicing_n = 5,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_short), "Short posts should be processed")

  # Test with long posts
  long_posts <- test_data[7:9, ]  # These are the long posts from our test data
  result_long <- parallel_execute(
    test_data = long_posts,
    slicing_n = 3,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_long), "Long posts should be processed")

  # Test 5: Different column configurations
  cat("\nTesting different column configurations...\n")

  # Create data with different column order
  test_data_reordered <- data.frame(
    some_other_col = "other",
    post_id = 1:8,
    full_post = test_data$full_post[1:8]
  )

  set_parameters(n_variables = 4, id_column = 2, text_column = 3, sep = ";")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  result_reordered <- parallel_execute(
    test_data = test_data_reordered,
    slicing_n = 8,
    cores = 2,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_reordered), "Reordered columns should work")

  # Test 6: Edge cases
  cat("\nTesting edge cases...\n")

  # Reset parameters for edge case test
  set_parameters(n_variables = 2, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract exactly 2 variables from each text: sentiment;confidence. Return ONLY the values separated by semicolons, no explanations or labels. Example: positive;0.8")

  # Test with very short slicing_n
  result_minimal <- parallel_execute(
    test_data = test_data[1:5, ],
    slicing_n = 3,
    cores = 1,
    multi_core = FALSE
  )

  expect_true(is.data.frame(result_minimal), "Minimal slicing should work")

  cat("\nFlexibility tests completed successfully!\n")
})

test_that("Error handling tests for rapidcodeR", {

  # Test 1: Invalid input data
  cat("\nTesting invalid input data...\n")

  # Test with non-dataframe input
  expect_error(
    parallel_execute(test_data = "not_a_dataframe", slicing_n = 10, cores = 2),
    "test_data must be a data frame"
  )

  # Set up basic parameters for error testing
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  # Test with empty dataframe
  empty_df <- data.frame()
  expect_error(
    parallel_execute(test_data = empty_df, slicing_n = 10, cores = 2),
    "id_column.*is greater than the number of columns"
  )

  # Test with missing required columns
  invalid_df <- data.frame(wrong_col = 1:10)
  expect_error(
    parallel_execute(test_data = invalid_df, slicing_n = 10, cores = 2),
    "text_column.*is greater than the number of columns"
  )

  # Test with NA values in text column
  df_with_na <- data.frame(
    post_id = 1:7,
    full_post = c("Valid text", NA, "Another valid text", "", "More text", "This is a longer valid text", "Another valid entry")
  )

  # This should work (NA values are filtered out)
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  # Skip this test if no API key
  if (nchar(Sys.getenv("OPENAI_API_KEY")) > 0) {
    result_with_na <- parallel_execute(
      test_data = df_with_na,
      slicing_n = 5,
      cores = 1
    )
    expect_true(is.data.frame(result_with_na), "Should handle NA values gracefully")
  }

  # Test 2: Invalid parameters
  cat("\nTesting invalid parameters...\n")

  # Set up API specs for parameter testing with real key
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  valid_df <- data.frame(
    post_id = 1:10,
    full_post = paste("Test post", 1:10)
  )

  # Test with negative slicing_n
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = -5, cores = 2),
    "slicing_n must be a positive number"
  )

  # Test with slicing_n greater than data rows
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 100, cores = 2),
    "Make sure that slicing_n is not greater than the number of rows"
  )

  # Test with zero cores
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = 0),
    "cores must be a positive number"
  )

  # Test with negative cores
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = -2),
    "cores must be a positive number"
  )

  # Test 3: Missing API configuration
  cat("\nTesting missing API configuration...\n")

  # Clear any existing API configuration
  if (exists("api_provider", envir = .package_env)) {
    rm("api_provider", envir = .package_env)
  }

  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = 2),
    "API provider not set"
  )

  # Test 4: Invalid API configuration
  cat("\nTesting invalid API configuration...\n")

  # Test invalid provider in set_api_specs
  expect_error(
    set_api_specs(provider = "InvalidProvider", model = "gpt-3.5-turbo", temp = 0.3,
                  api_key = Sys.getenv("OPENAI_API_KEY")),
    "provider must be either 'OpenAI' or 'Groq'"
  )

  # Test 5: Missing coding instruction (skip on CI when no API key: API test runs first and fails)
  cat("\nTesting missing coding instruction...\n")
  testthat::skip_if(
    nchar(Sys.getenv("OPENAI_API_KEY")) == 0,
    "OpenAI API key not set; cannot test coding instruction error path"
  )

  # Set valid API config but clear coding instruction
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  if (exists("coding_instruction", envir = .package_env)) {
    rm("coding_instruction", envir = .package_env)
  }

  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = 2),
    "Coding instruction has not been set"
  )

  # Test 6: Invalid set_parameters calls
  cat("\nTesting invalid set_parameters calls...\n")

  # Test with invalid n_variables
  expect_error(
    set_parameters(n_variables = -1),
    "n_variables must be a positive number between 1 and 20"
  )

  expect_error(
    set_parameters(n_variables = 25),
    "n_variables must be a positive number between 1 and 20"
  )

  # Test with invalid id_column
  expect_error(
    set_parameters(id_column = 0),
    "id_column must be a positive number"
  )

  expect_error(
    set_parameters(id_column = -2),
    "id_column must be a positive number"
  )

  # Test with invalid text_column
  expect_error(
    set_parameters(text_column = 0),
    "text_column must be a positive number"
  )

  # Test with invalid separator
  expect_error(
    set_parameters(sep = "invalid"),
    "sep must be one of:"
  )

  # Test 7: Invalid set_api_specs calls
  cat("\nTesting invalid set_api_specs calls...\n")

  # Test with invalid provider
  expect_error(
    set_api_specs(provider = "Invalid", model = "gpt-3.5-turbo", temp = 0.3, api_key = "key"),
    "provider must be either 'OpenAI' or 'Groq'"
  )

  # Test with invalid temperature
  expect_error(
    set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 1.5, api_key = "key"),
    "temp must be a single numeric value between 0 and 1"
  )

  expect_error(
    set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = -0.1, api_key = "key"),
    "temp must be a single numeric value between 0 and 1"
  )

  # Test 8: Invalid seed parameter
  cat("\nTesting invalid seed parameter...\n")

  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3, api_key = Sys.getenv("OPENAI_API_KEY"))
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")

  # Test with non-integer seed
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = 2, seed = 1.5),
    "seed must be NULL or a single integer value"
  )

  # Test with vector seed
  expect_error(
    parallel_execute(test_data = valid_df, slicing_n = 10, cores = 2, seed = c(1, 2)),
    "seed must be NULL or a single integer value"
  )

  # Test 9: Text column validation
  cat("\nTesting text column validation...\n")

  # Test with non-character text column
  df_numeric_text <- data.frame(
    post_id = 1:5,
    full_post = 1:5  # Numeric instead of character
  )

  expect_error(
    parallel_execute(test_data = df_numeric_text, slicing_n = 5, cores = 2),
    "Text column.*must contain character or factor data"
  )

  # Test with all NA text column
  df_all_na_text <- data.frame(
    post_id = 1:5,
    full_post = rep(NA_character_, 5)
  )

  expect_error(
    parallel_execute(test_data = df_all_na_text, slicing_n = 5, cores = 2),
    "Text column.*contains only NA values"
  )

  # Test 10: Core count warnings
  cat("\nTesting core count warnings...\n")

  # Ensure parameters are set before parallel_execute
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract: sentiment,confidence,topic,relevance")
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  # This should produce a warning but not an error
  expect_warning(
    parallel_execute(test_data = valid_df, slicing_n = 5, cores = 20),
    "Requested cores.*exceeds available cores"
  )

  # API key is already set to real key throughout the test


  cat("\nError handling tests completed successfully!\n")
})

test_that("Performance tests for rapidcodeR", {

  # Skip if API keys are not set
  skip_if_not(is_api_key_available(), "OpenAI API key not set")

  # Test API connectivity first
  set_api_specs(provider = "OpenAI", model = "gpt-3.5-turbo", temp = 0.3,
                api_key = Sys.getenv("OPENAI_API_KEY"))

  # Try to make a simple API call to test connectivity
  test_result <- tryCatch({
    ask_openai("What is the capital of Norway?")
  }, error = function(e) {
    return(NA)
  })

  # Skip test if API connection fails
  skip_if(is.na(test_result), "OpenAI API test failed - check API key and internet connection")

  # Create test data
  test_data <- data.frame(
    post_id = 1:50,
    full_post = paste("This is test post number", 1:50,
                     "about various topics including technology, science, and general discussion.")
  )

  # Set up package configuration
  set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")
  set_coding_instruction("Extract exactly 4 variables from each text: sentiment;confidence;topic;relevance. Return ONLY the values separated by semicolons, no explanations or labels. Example: positive;0.8;technology;high")

  # Test 1: Small dataset performance (10 posts)
  cat("\nTesting small dataset performance (10 posts)...\n")
  start_time <- Sys.time()

  result_small <- parallel_execute(
    test_data = test_data[1:10, ],
    slicing_n = 10,
    cores = 2,
    multi_core = FALSE
  )

  small_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
  cat("Small dataset processing time:", small_time, "seconds\n")

  # Skip if parallel processing failed (can happen in check() environment)
  skip_if(nrow(result_small) == 0, "Parallel processing failed - likely environment issue in check()")

  # Performance expectations
  expect_true(small_time < 120, "Small dataset should complete within 2 minutes")
  expect_true(nrow(result_small) > 0, "Small dataset should produce results")

  # Test 2: Medium dataset performance (30 posts)
  cat("\nTesting medium dataset performance (30 posts)...\n")
  start_time <- Sys.time()

  result_medium <- parallel_execute(
    test_data = test_data[1:30, ],
    slicing_n = 30,
    cores = 4,
    multi_core = FALSE
  )

  medium_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
  cat("Medium dataset processing time:", medium_time, "seconds\n")

  # Skip if parallel processing failed (can happen in check() environment)
  skip_if(nrow(result_medium) == 0, "Parallel processing failed - likely environment issue in check()")

  # Performance expectations
  if (medium_time >= 300) {
    warning("Medium dataset took ", round(medium_time, 1), " seconds (", round(medium_time/60, 1), " minutes) - ideally should complete within 5 minutes")
  }
  expect_true(nrow(result_medium) > 0, "Medium dataset should produce results")

  # Test 3: Core scaling performance
  cat("\nTesting core scaling performance...\n")

  # Test with 2 cores - use smaller slicing_n to create multiple batches
  cat("Starting 2-core test (2 batches)...\n")
  start_time_2cores <- Sys.time()
  result_2cores <- parallel_execute(
    test_data = test_data[1:20, ],
    slicing_n = 10,  # Creates 2 batches for 2 cores
    cores = 2,
    multi_core = FALSE
  )
  time_2cores <- as.numeric(difftime(Sys.time(), start_time_2cores, units = "secs"))
  cat("2 cores completed in", round(time_2cores, 2), "seconds. Rows processed:", nrow(result_2cores), "\n")

  # Test with 4 cores (if available)
  if (parallel::detectCores() >= 4) {
    cat("Starting 4-core test (4 batches)...\n")
    start_time_4cores <- Sys.time()
    result_4cores <- parallel_execute(
      test_data = test_data[1:40, ],  # More data for 4 cores
      slicing_n = 10,  # Creates 4 batches for 4 cores
      cores = 4,
      multi_core = FALSE
    )
    time_4cores <- as.numeric(difftime(Sys.time(), start_time_4cores, units = "secs"))
    cat("4 cores completed in", round(time_4cores, 2), "seconds. Rows processed:", nrow(result_4cores), "\n")

    # Check if 4 cores is significantly slower - with proper batching it shouldn't be
    # Allow some overhead but expect better or similar performance
    if (time_4cores >= time_2cores * 2.0) {
      warning(paste0("Performance issue detected: 4 cores (", round(time_4cores, 2),
                    "s) is significantly slower than 2 cores (", round(time_2cores, 2),
                    "s). This may be due to API rate limiting, worker overhead, or resource contention."))
    } else {
      cat("Core scaling looks reasonable. 4 cores:", round(time_4cores, 2), "s vs 2 cores:", round(time_2cores, 2), "s\n")
    }
  }

  # Test 4: Memory usage monitoring
  cat("\nTesting memory usage...\n")
  memory_before <- gc()["Vcells", "used"]

  result_memory <- parallel_execute(
    test_data = test_data[1:25, ],
    slicing_n = 25,
    cores = 2,
    multi_core = FALSE
  )

  memory_after <- gc()["Vcells", "used"]
  memory_increase <- memory_after - memory_before

  cat("Memory increase:", memory_increase, "Vcells\n")
  # Allow reasonable memory increase - parallel processing uses more memory
  expect_true(memory_increase < 500000, "Memory increase should be reasonable")

  # Test 5: Result structure validation
  expect_true(is.data.frame(result_small), "Result should be a data frame")

  expect_true(ncol(result_small) == 4, "Result should have 4 variables")
  expect_true(all(c("Var1", "Var2", "Var3", "Var4") %in% names(result_small)),
              "Result should have correct column names")

  cat("\nPerformance tests completed successfully!\n")
})

#' Process Text Data Using OpenAI GPT API
#'
#' This function sends text data to the OpenAI GPT API for variable extraction,
#' optimized for speed and cost efficiency. It formats the data for API consumption
#' and parses the CSV-formatted response with robust error handling.
#'
#' @importFrom stats setNames
#' @importFrom stringr str_remove
#'
#' @param data_subset Data frame. Subset of text data to process. Must contain
#'   columns specified by `id_column` and `text_column` parameters (set via `set_parameters()`).
#' @param n_post Integer. Number of posts to process in this batch.
#' @param coding_instruction Character. Instructions for the AI model specifying
#'   how to extract and format variables from the text.
#' @param worker_env Environment or NULL. When running in a parallel worker, the worker's
#'   package environment. If NULL (default), \code{.package_env} is used.
#'
#' @details
#' The function performs several optimizations:
#' \itemize{
#'   \item Filters out very short posts (< 4 characters) to reduce noise
#'   \item Sanitizes text by replacing problematic characters (semicolons, quotes)
#'   \item Constructs a single prompt with multiple posts for batch processing
#'   \item Calls OpenAI API and handles errors gracefully
#'   \item Parses the response assuming CSV format with newline separators
#'   \item Always returns a dataframe (single rows become 1-row dataframes)
#' }
#'
#' The function includes specific error handling for API failures and will return
#' NULL if the API returns an error or invalid response.
#'
#' @return Data frame. Each row contains the AI's response for one post,
#'   with columns named Var1, Var2, etc. Returns NULL if no posts remain after filtering
#'   or if API errors occur.
#'
#' @seealso [ask_openai()], [groq_func()]
#' @keywords internal

gpt_func <- function(data_subset, n_post, coding_instruction, worker_env = NULL) {

  env <- if (is.null(worker_env)) .package_env else worker_env

  # Use all rows in the pre-sampled data subset
  posts <- data_subset

  initial_count <- nrow(posts)

  # Get column parameters from package environment
  id_column <- get("id_column", envir = env)
  text_column <- get("text_column", envir = env)

  # Get column names for easier reference (data_subset has internal_id already removed by main_func)
  id_col_name <- names(posts)[id_column]
  text_col_name <- names(posts)[text_column]

  #Remove ultra-short blurbs:
  posts <- subset(posts,nchar(posts[[text_col_name]]) >= 4)

  filtered_count <- nrow(posts)
  if (initial_count != filtered_count) {
    message("Filtered out ", initial_count - filtered_count, " posts with < 4 characters")
  }

  if (nrow(posts) == 0) {
    message("No posts left to process, breaking loop.")
    return(NULL)
  }

  # Get expected number of variables and separator
  n_variables <- if (exists("n_variables", envir = env)) {
    get("n_variables", envir = env)
  } else {
    9  # Default fallback
  }

  sep <- if (exists("sep", envir = env)) {
    get("sep", envir = env)
  } else {
    ";"  # Default fallback
  }

  # Sanitize text content - replace the separator with a colon to avoid conflicts
  posts[[text_col_name]] <- gsub(sep, ":", posts[[text_col_name]])
  posts[[text_col_name]] <- gsub("'", "`", posts[[text_col_name]])
  posts[[text_col_name]] <- gsub("\"", "`", posts[[text_col_name]])

  #Build GPT query out of posts using dynamic column names:
  gpt_posts <- ""
  for (i in seq_len(nrow(posts))){
    # Always include the ID column in the prompt
    gpt_posts <- paste(gpt_posts, posts[[id_col_name]][i], sep, posts[[text_col_name]][i])
  }

  # Extract the temperature and model name from package environment
  api_model <- get("api_model", envir = env)
  api_temp <- get("api_temp", envir = env)

  gpt_prompt <- paste(coding_instruction, gpt_posts)

  gpt_response <- NA

  gpt_response <- ask_openai(gpt_prompt, temp = api_temp, model = api_model)

  if (is.na(gpt_response) || gpt_response == "API_ERROR") {
    message("Skipping due to API error.")
    return(NULL)
  }

  # Parse multiple lines from API response
  codings <- unlist(strsplit(gpt_response,"\n"))

  # Remove trailing separators and empty lines
  codings <- str_remove(codings, paste0("\\", sep, "\\s*$"))
  codings <- codings[codings != ""]

  # If no valid codings, return NULL
  if (length(codings) == 0 || all(is.na(codings))) {
    return(NULL)
  }

  # Process each coding line into a dataframe row
  result_rows <- list()

  for (i in seq_along(codings)) {
    if (!is.na(codings[i]) && codings[i] != "") {
      # Parse the coding line using the dynamic separator
      values <- unlist(strsplit(codings[i], sep))
      values <- trimws(values)

      # Ensure we have the expected number of columns
      while(length(values) < n_variables) {
        values <- c(values, NA)
      }

      if (length(values) >= n_variables) {
        # Create column names
        col_names <- paste0("Var", 1:n_variables)

        # Create a single-row dataframe
        row_df <- as.data.frame(
          as.list(setNames(values[1:n_variables], col_names)),
          stringsAsFactors = FALSE
        )

        result_rows[[length(result_rows) + 1]] <- row_df
      }
    }
  }

  # Combine all rows into a single dataframe
  if (length(result_rows) > 0) {
    result_df <- do.call(bind_rows, result_rows)
    return(result_df)
  } else {
    return(NULL)
  }

}

#' Track Progress of Parallel Processing Tasks
#'
#' This function wraps the main processing function with progress tracking,
#' error handling, and rate limiting. It's designed to be called from parallel
#' workers to process data batches while providing feedback on progress.
#'
#' @param df Data frame. A subset of data to process in this task.
#' @param index Integer. The current task number (for progress tracking).
#' @param total_tasks Integer. Total number of tasks in the parallel job.
#' @param n_post Integer. Number of posts to process in each API call.
#' @param provider Character. AI provider to use, either "OpenAI" or "Groq".
#' @param worker_env Environment or NULL. When running in a parallel worker, the worker's
#'   package environment so config (e.g. \code{coding_instruction}, \code{id_column})
#'   is available. If NULL (default), the package's \code{.package_env} is used.
#' @param verbose Logical. If TRUE (default), progress messages are shown. If FALSE, suppressed.
#'
#' @details
#' The function provides several important features:
#' \itemize{
#'   \item Progress messages showing current task vs total tasks
#'   \item Rate limiting with 5-second delays to avoid overwhelming APIs
#'   \item Comprehensive error handling with informative error messages
#'   \item Automatic validation of provider parameter
#'   \item Graceful degradation returning NULL on errors
#' }
#'
#' This function is primarily used internally by the parallel processing
#' framework and calls `main_func()` with the `n_post` parameter passed by the caller.
#'
#' @return Data frame with successfully processed results from `main_func()`,
#'   or NULL if an error occurs during processing.
#'
#' @seealso [main_func()], [parallel_execute()]
#' @keywords internal
track_progress <- function(df, index, total_tasks, n_post, batch_size, provider, worker_env = NULL, verbose = TRUE) {

  # Input validation
  if (!is.data.frame(df)) {
    stop("df must be a data frame")
  }

  if (!is.numeric(index) || index <= 0) {
    stop("index must be a positive number")
  }

  if (!is.numeric(total_tasks) || total_tasks <= 0) {
    stop("total_tasks must be a positive number")
  }

  if (!is.numeric(batch_size) || batch_size <= 0) {
    stop("batch_size must be a positive number")
  }

  if (!provider %in% c("OpenAI", "Groq")) {
    stop("provider must be either 'OpenAI' or 'Groq'")
  }

  tryCatch({
    if (verbose) message("Processing task ", index, " out of ", total_tasks)
    Sys.sleep(5)  # 5-second delay to avoid hitting the API too hard
    if(provider %in% c("OpenAI", "Groq")) {
      to_code_max_id <- nrow(df)
      result <- main_func(df, to_code_max_id, n_post, provider, worker_env = worker_env)
    }
    else {
      stop("Please specify whether provider is 'OpenAI' or 'Groq'")
    }
    if (verbose) message("Finished task ", index)
    return(result)
  }, error = function(e) {
    message("Error processing task ", index, ": ", e$message)
    return(NULL)  # Return NULL or handle the error appropriately
  })
}

#' Calculate Inter-Rater Agreement Across Multiple Datasets
#'
#' This function calculates the percentage overlap (agreement) between multiple
#' datasets containing the same variables. It's designed to assess reliability
#' and consistency when the same data is processed multiple times or by different
#' systems/raters.
#'
#' @param datasets List. A list of data frames to compare. Each data frame should
#'   have the same column structure with variable columns (Var1, Var2, etc.).
#'
#' @param alpha Logical. If TRUE, also compute Krippendorff's Alpha (nominal)
#'   for each variable across the provided datasets. Defaults to FALSE.
#'
#' @param key_cols Character vector or NULL. Column name(s) used to match rows
#'   across datasets. Use this when datasets have different row counts due to a
#'   compound row identity (e.g. \code{key_cols = c("Var1", "Var3")}). Rows with
#'   NA in any key column are dropped before matching. When NULL (default), rows
#'   are matched on the first column only, which is the standard behaviour for
#'   single-response output from \code{parallel_execute()}.
#'
#' @details
#' The function performs comprehensive overlap analysis:
#' \itemize{
#'   \item Finds common rows across all datasets based on \code{key_cols}
#'   \item Aligns and sorts datasets by the key before comparing
#'   \item Calculates pairwise agreement for all possible dataset combinations
#'   \item Computes variable-wise agreement percentages
#'   \item Returns average agreement across all dataset pairs
#' }
#'
#' Agreement is calculated as the percentage of cases where values match exactly.
#' The number of rows used in the comparison is printed as a message (\code{N = ...}).
#'
#' @return Matrix. If alpha = FALSE: one-row matrix with average percentage
#'   agreement for each variable (columns), rounded to 2 decimals.
#'   If alpha = TRUE: two-row matrix with row "Overlap" (percent agreement)
#'   and row "Alpha" (Krippendorff's Alpha for nominal data), rounded
#'   to 3 decimals for Alpha.
#'
#' @examples
#' \dontrun{
#' # Standard use: compare two processing runs (single-response output)
#' run1 <- parallel_execute(my_data, slicing_n = 100, cores = 4)
#' run2 <- parallel_execute(my_data, slicing_n = 100, cores = 4)
#' calculate_overlap(list(run1, run2))
#'
#' # Multi-response output: datasets may differ in row count;
#' # match on compound key (e.g. speech ID + paragraph number)
#' calculate_overlap(list(run1, run2), key_cols = c("Var1", "Var3"))
#' }
#'
#' @seealso [parallel_execute()]
#' @importFrom stats complete.cases
#' @export
calculate_overlap <- function(datasets, alpha = FALSE, key_cols = NULL) {

  # Input validation
  if (!is.list(datasets) || length(datasets) < 2) {
    stop("datasets must be a list with at least 2 data frames")
  }
  if (!all(sapply(datasets, is.data.frame))) {
    stop("All elements in datasets must be data frames")
  }

  num_datasets <- length(datasets)
  first_names <- names(datasets[[1]])

  if (!all(sapply(datasets, function(df) identical(names(df), first_names)))) {
    stop("All datasets must have the same column names.")
  }

  if (!is.null(key_cols)) {
    if (!is.character(key_cols) || !all(key_cols %in% first_names)) {
      stop("key_cols must be a character vector of column names present in all datasets.")
    }
  }

  # Align datasets on common rows -----------------------------------------------

  if (!is.null(key_cols)) {

    # Compound key: paste key columns together, drop rows with NA in any key col
    make_key <- function(df) {
      key_df <- df[, key_cols, drop = FALSE]
      complete <- complete.cases(key_df)
      df_clean <- df[complete, ]
      keys <- do.call(paste, c(df_clean[, key_cols, drop = FALSE], sep = "|||"))
      list(df = df_clean, keys = keys)
    }

    keyed <- lapply(datasets, make_key)
    common_keys <- Reduce(intersect, lapply(keyed, `[[`, "keys"))

    if (length(common_keys) == 0) {
      stop("No common rows found across datasets using the specified key_cols.")
    }

    datasets <- lapply(keyed, function(x) {
      matched <- x$df[x$keys %in% common_keys, ]
      sort_order <- order(do.call(paste, c(matched[, key_cols, drop = FALSE], sep = "|||")))
      matched <- matched[sort_order, ]
      matched[!duplicated(matched[, key_cols, drop = FALSE]), ]
    })

  } else {

    # Default: align on first column only
    common_ids <- Reduce(intersect, lapply(datasets, function(df) df[[1]]))

    datasets <- lapply(datasets, function(df) {
      filtered <- df[df[[1]] %in% common_ids, ]
      sort_order <- order(as.numeric(filtered[[1]]))
      filtered <- filtered[sort_order, ]
      filtered[!duplicated(filtered), ]
    })
  }

  # Verify alignment succeeded
  row_counts <- sapply(datasets, nrow)
  if (length(unique(row_counts)) > 1) {
    stop(
      "Datasets have different row counts after alignment (",
      paste(row_counts, collapse = ", "), " rows). ",
      "If your data uses a compound row key, specify key_cols."
    )
  }

  # Pairwise overlap ------------------------------------------------------------

  n_pairs <- num_datasets * (num_datasets - 1L) / 2L
  overlaps_list <- vector("list", n_pairs)
  all_vars <- first_names
  pair_index <- 1L

  for (i in 1:(num_datasets - 1)) {
    for (j in (i + 1):num_datasets) {
      overlaps_list[[pair_index]] <- sapply(all_vars, function(var) {
        mean(datasets[[i]][[var]] == datasets[[j]][[var]], na.rm = TRUE)
      })
      pair_index <- pair_index + 1L
    }
  }

  avg_overlap <- round(
    (Reduce("+", overlaps_list) / n_pairs) * 100,
    digits = 2
  )

  n_rows <- nrow(datasets[[1]])
  message("N = ", n_rows)

  result_mat <- matrix(
    avg_overlap, nrow = 1, byrow = TRUE,
    dimnames = list("Overlap", all_vars)
  )

  if (isTRUE(alpha)) {
    alpha_vals <- sapply(all_vars, function(var) {
      ratings <- do.call(rbind, lapply(datasets, function(df) as.character(df[[var]])))
      if (!requireNamespace("irr", quietly = TRUE)) return(NA_real_)
      kripp_alpha <- get("kripp.alpha", asNamespace("irr"))
      suppressWarnings(out <- try(kripp_alpha(ratings, method = "nominal"), silent = TRUE))
      if (inherits(out, "try-error")) NA_real_ else as.numeric(round(out$value, digits = 3))
    })
    result_mat <- rbind(result_mat, Alpha = alpha_vals)
  }

  return(result_mat)
}

# Package imports are handled by DESCRIPTION Imports field

#' Send Prompt to GROQ AI API
#'
#' This function sends a text prompt to the GROQ AI API and returns the model's
#' response. It includes error handling and uses the groqR package for API communication.
#'
#' @param prompt Character. The text prompt to send to the AI model.
#' @param temp Numeric. Temperature parameter controlling randomness (0-1).
#'   Lower values make output more deterministic. Required (typically from set_api_specs()).
#' @param topp Numeric. Top-p parameter controlling diversity (0-1). Default is 1.
#' @param model Character. GROQ model to use. Required (typically from set_api_specs()).
#' @param api_key Character. GROQ API key. Must be provided.

#'
#' @details
#' The function uses the groqR package to communicate with GROQ's API. It includes
#' error handling that returns NA if the API call fails. The function is optimized
#' for text processing tasks with a maximum token limit of 5000.
#'
#' @return Character. The AI model's response to the prompt, or NA if an error occurs.
#'
#' @examples
#' \dontrun{
#' # Set API key first
#' Sys.setenv(GROQ_API_KEY = "your_api_key")
#'
#' # Send a simple prompt
#' response <- ask_groq("What is the capital of France?")
#'
#' # Use different parameters
#' response <- ask_groq("Analyze this text", temp = 0.3, topp = 0.9)
#' }
#'
#' @seealso [ask_openai()]
#' @export
ask_groq <- function(prompt, temp, topp = 1, model, api_key) {

  # Input validation
  if (!is.character(prompt) || length(prompt) != 1) {
    stop("prompt must be a single character string")
  }

  # Get model and temp from set_api_specs if not provided
  if (is.null(model)) {
    if (exists("api_model", envir = globalenv())) {
      model <- get("api_model", envir = globalenv())
    } else if (exists("api_model", envir = .package_env)) {
      model <- get("api_model", envir = .package_env)
    } else {
      stop("Model not set. Please use set_api_specs() to configure your API settings.")
    }
  }

  if (is.null(temp)) {
    if (exists("api_temp", envir = globalenv())) {
      temp <- get("api_temp", envir = globalenv())
    } else if (exists("api_temp", envir = .package_env)) {
      temp <- get("api_temp", envir = .package_env)
    } else {
      stop("Temperature not set. Please use set_api_specs() to configure your API settings.")
    }
  }

  if (!is.numeric(temp) || temp < 0 || temp > 1) {
    stop("temp must be a number between 0 and 1")
  }

  if (!is.numeric(topp) || topp < 0 || topp > 1) {
    stop("topp must be a number between 0 and 1")
  }

  if (!is.character(model) || length(model) != 1) {
    stop("model must be a single character string")
  }

  if (identical(api_key, "")) {
    stop("Environment variable 'GROQ_API_KEY' is not set. Please set it with Sys.setenv() or via your .Renviron file.")
  }

  result <- tryCatch({
    # Ensure groqR package is loaded
    if (!requireNamespace("groqR", quietly = TRUE)) {
      stop("groqR package is required but not installed")
    }

    # Set the API key as an environment variable for groqR
    Sys.setenv(GROQ_API_KEY = api_key)

    groqR::ask(
      prompt,
      systemRole   = "Your assistant",
      model        = model,
      maxTokens    = getOption("rapidcodeR.max_tokens", 5000),
      temperature  = temp,
      top_p        = topp,
      proxy        = NULL
    )
  }, error = function(e) {
    warning("Got API error: ", e$message)
    NA
  })

  return(result)
}

#' Send Prompt to OpenAI API
#'
#' This function sends a text prompt to the OpenAI API and returns the model's
#' response. It includes robust error handling and uses the openai package for API communication.
#'
#' @param prompt Character. The text prompt to send to the AI model.
#' @param temp Numeric. Temperature parameter controlling randomness (0-1).
#'   Lower values make output more deterministic. If NULL, retrieved from set_api_specs().
#' @param topp Numeric. Top-p parameter controlling diversity (0-1). Default is 1.
#' @param model Character. OpenAI model to use. If NULL, retrieved from set_api_specs().
#'
#' @details
#' The function uses the openai package to communicate with OpenAI's API. It includes
#' comprehensive error handling that returns NA if the API call fails or returns
#' an unexpected format. The function expects the OPENAI_API_KEY environment variable
#' to be set via `set_api_specs()`.
#'
#' @return Character. The AI model's response to the prompt, or NA if an error occurs.
#'
#' @examples
#' \dontrun{
#' # Set API specifications first
#' set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_api_key")
#'
#' # Send a simple prompt
#' response <- ask_openai("What is the capital of France?")
#'
#' # Use different parameters
#' response <- ask_openai("Analyze this text", temp = 0.3, topp = 0.9)
#' }
#'
#' @seealso [ask_groq()], [set_api_specs()]
#' @export
ask_openai <- function(prompt, temp = NULL, topp = 1, model = NULL) {

  # Input validation
  if (!is.character(prompt) || length(prompt) != 1) {
    stop("prompt must be a single character string")
  }

  # Get model and temp from set_api_specs if not provided
  if (is.null(model)) {
    if (exists("api_model", envir = globalenv())) {
      model <- get("api_model", envir = globalenv())
    } else if (exists("api_model", envir = .package_env)) {
      model <- get("api_model", envir = .package_env)
    } else {
      stop("Model not set. Please use set_api_specs() to configure your API settings.")
    }
  }

  if (is.null(temp)) {
    if (exists("api_temp", envir = globalenv())) {
      temp <- get("api_temp", envir = globalenv())
    } else if (exists("api_temp", envir = .package_env)) {
      temp <- get("api_temp", envir = .package_env)
    } else {
      stop("Temperature not set. Please use set_api_specs() to configure your API settings.")
    }
  }

  if (!is.numeric(temp) || temp < 0 || temp > 1) {
    stop("temp must be a number between 0 and 1")
  }

  if (!is.numeric(topp) || topp < 0 || topp > 1) {
    stop("topp must be a number between 0 and 1")
  }

  if (!is.character(model) || length(model) != 1) {
    stop("model must be a single character string")
  }

  result <- tryCatch({
    # Ensure openai package is loaded
    if (!requireNamespace("openai", quietly = TRUE)) {
      stop("openai package is required but not installed")
    }

    response <- openai::create_chat_completion(
      model = model,
      messages = list(
        list(role = "system", content = "You are a helpful assistant."),
        list(role = "user", content = prompt)
      ),
      temperature = temp,
      top_p = topp
    )


    # Handle response$choices as either a data.frame or list
    if (!is.null(response$choices)) {
      if (is.data.frame(response$choices)) {
        # If it's a data.frame, get the first row's message content
        if (nrow(response$choices) > 0) {
          # Check for message.content column (new format)
          if ("message.content" %in% names(response$choices)) {
            content <- response$choices[1, "message.content"]
            if (is.character(content) && !is.na(content)) {
              return(content)
            }
          }
          # Check for nested message structure (old format)
          else if ("message" %in% names(response$choices)) {
            first_choice <- response$choices[1, "message"]
            if (is.list(first_choice) && "content" %in% names(first_choice)) {
              return(first_choice$content)
            } else if (is.character(first_choice)) {
              return(first_choice)
            }
          }
        }
      } else if (is.list(response$choices) && length(response$choices) > 0) {
        # If it's a list, get the first element's message content
        first_choice <- response$choices[[1]]
        if (is.list(first_choice) && "message" %in% names(first_choice)) {
          message_obj <- first_choice$message
          if (is.list(message_obj) && "content" %in% names(message_obj)) {
            return(message_obj$content)
          }
        }
      }
    }

    message("Unexpected API response format.")
    return(NA)

  }, error = function(e) {
    message("API error: ", e$message)
    return(NA)
  })

  return(result)
}

#' Parse AI Response into Formatted Row String
#'
#' This function takes a single AI response string and formats it into a standardized
#' SQL-style row format for consistent data processing. It handles text sanitization
#' and ensures the response has exactly N elements (N set by set_parameters()).
#'
#' @param ai_response Character. A single response string from an AI model, typically
#'   containing semicolon-separated values.
#'
#' @details
#' The function performs several formatting operations:
#' \itemize{
#'   \item Converts input to character format
#'   \item Replaces quotes with backticks to avoid SQL conflicts
#'   \item Splits on semicolons to extract individual values
#'   \item Trims whitespace from each value
#'   \item Pads with NA values if fewer than expected elements
#'   \item Validates that we have the expected number of elements
#'   \item Converts to lowercase and formats as SQL-style row
#' }
#'
#' The function expects exactly N values as set by set_parameters().
#'
#' @return Character. A formatted row string in SQL format like "('val1','val2',...)",
#'   or NA if validation fails.
#'
#' @examples
#' \dontrun{
#' # AI response with semicolon-separated values (assuming set_parameters(n_variables = 6))
#' ai_response <- "123;en;Hello world;positive;high;topic1"
#' formatted_row <- make_value_row(ai_response)
#' # Returns: "('123','en','hello world','positive','high','topic1')"
#' }
#'
#' @seealso [main_func()]
#' @export
make_value_row <- function(ai_response) {
  # Input validation
  if (!is.character(ai_response) || length(ai_response) != 1) {
    stop("ai_response must be a single character string")
  }

  if (is.na(ai_response) || ai_response == "") {
    return(NA)
  }

  # No longer need to enforce a specific number of columns
  # The function will handle any number of columns dynamically

  separator <- if (exists("sep", envir = globalenv())) {
    get("sep", envir = globalenv())
  } else {
    ";" # Default fallback
  }

  # Ensure input is a single string
  ai_response <- as.character(ai_response)

  # Replace single/double quotes with backticks
  ai_response <- gsub("['\"]", "`", ai_response)

  # Split into values based on semicolons
  insert_values <- unlist(strsplit(ai_response, split = separator))

  # Trim whitespace from each entry
  insert_values <- trimws(insert_values)

  # Get expected number of variables from package environment
  n_variables <- if (exists("n_variables", envir = globalenv())) {
    get("n_variables", envir = globalenv())
  } else {
    9  # Default fallback (matches set_parameters default)
  }

  # Ensure we have the expected number of columns
  while(length(insert_values) < n_variables) {
    insert_values = c(insert_values, NA)
  }

  # Ensure we have the expected number of elements
  if (length(insert_values) == n_variables) {
    # Format each element safely (escape single quotes for SQL-style output)
    insert_values <- gsub("'", "''", insert_values)
    insert_values <- tolower(insert_values)  # Convert to lowercase (optional)

    # Build SQL-style row string
    insert_row <- paste0("('", paste(insert_values, collapse = "','"), "')")
  } else {
    warning(paste("make_value_row: Expected", n_variables, "values, got", length(insert_values),
                  "values:", paste(insert_values, collapse = ", ")))
    insert_row <- NA
  }

  return(insert_row)
}

# Test helper functions
# This file is automatically sourced by testthat before running tests

# Function to get API key from various sources
get_test_api_key <- function() {
  # Try different sources in order of preference
  sources <- c(
    Sys.getenv("OPENAI_API_KEY"),
    Sys.getenv("RAPIDCODER_TEST_API_KEY"),
    # Add other sources as needed
    ""
  )
  
  # Return first non-empty key
  for (key in sources) {
    if (nchar(key) > 0) {
      return(key)
    }
  }
  
  return("")
}

# Test if API key is available
is_api_key_available <- function() {
  nchar(get_test_api_key()) > 0
}

# Get API key for tests
get_api_key <- function() {
  key <- get_test_api_key()
  if (nchar(key) == 0) {
    stop("No API key found. Set OPENAI_API_KEY or RAPIDCODER_TEST_API_KEY environment variable.")
  }
  return(key)
}

#' Execute Parallel Text Processing with AI Models
#'
#' This is the main function that orchestrates parallel processing of text data
#' using AI APIs for variable extraction. It optimizes for speed and cost by
#' distributing work across multiple cores and handling missing data efficiently.
#'
#' @import future
#' @import future.apply
#' @importFrom dplyr filter slice_sample arrange distinct bind_rows
#' @importFrom dplyr "%>%" .data
#' @importFrom rlang sym
#'
#' @param test_data Data frame. Input data containing text to be processed.
#' @param slicing_n Integer. Number of rows to sample from the input data for processing.
#'   The batch size will be automatically calculated by dividing this by the number of cores.
#' @param n_post Integer. Number of posts to process in each API call. Default is 15.
#'   Must be between 1 and 1000. Larger values may be more efficient but use more API credits.
#' @param cores Integer. Number of CPU cores to use for parallel processing. Default is 8.
#'   The function will use the provider specified in set_api_specs() to determine
#'   whether to use OpenAI or GROQ services.
#' @param seed Integer or NULL. Random seed for reproducible sampling. If NULL (default),
#'   sampling is random. If set to a number, ensures identical datasets across runs.
#' @param multi_core Logical. If TRUE, uses multicore backend for parallel processing
#'   (faster but Unix/Mac only). If FALSE (default), uses multisession backend
#'   (works on all platforms including Windows).
#' @param benchmarking Logical. If TRUE, returns processing time in seconds instead of
#'   the result data frame. Default is FALSE.
#' @param verbose Logical. If TRUE (default), progress and status messages are
#'   printed via \code{message()}. If FALSE, such messages are suppressed.
#'
#' @details
#' The function implements a multi-stage parallel processing workflow:
#' \enumerate{
#'   \item Validates parameters and tests API connectivity
#'   \item Randomly samples data and splits into batches for parallel processing
#'   \item Executes parallel processing using futures and tracks progress
#'   \item Handles missing/failed extractions with a secondary processing stage
#'   \item Combines and deduplicates results
#' }
#'
#' The function requires a coding instruction to be set via `set_coding_instruction()`
#' before execution. This instruction tells the AI how to format its responses.
#'
#' @return Data frame. Processed results containing extracted variables,
#'   with duplicates removed (when \code{multi_response=FALSE}) and sorted by ID.
#'
#' @examples
#' \dontrun{
#' # Set up API specifications and coding instruction
#' set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_key")
#' set_coding_instruction("Extract sentiment: id,sentiment,confidence")
#'
#' # Process data with random sampling (default)
#' results <- parallel_execute(
#'   test_data = my_data,
#'   slicing_n = 240,
#'   cores = 4
#' )
#'
#' # Process data with reproducible sampling (for calculate_overlap)
#' results1 <- parallel_execute(
#'   test_data = my_data,
#'   slicing_n = 240,
#'   cores = 4,
#'   seed = 123
#' )
#'
#' # Switch to GROQ for comparison
#' set_api_specs(provider = "Groq", model = "llama-3.3-70b-versatile",
#'   temp = 0.3, api_key = "your_groq_key")
#' results2 <- parallel_execute(
#'   test_data = my_data,
#'   slicing_n = 240,
#'   cores = 4,
#'   seed = 123
#' )
#'
#' # Calculate overlap between identical datasets
#' overlap_scores <- calculate_overlap(list(results1, results2))
#'
#' }
#'
#' @seealso [set_api_specs()], [set_coding_instruction()], [set_parameters()]
#' @export
parallel_execute <- function(test_data,
                             slicing_n,
                             n_post = 15,
                             cores = 8,
                             seed = NULL,
                             multi_core = FALSE,
                             benchmarking = FALSE,
                             verbose = TRUE) {

  # Input validation
  if (!is.data.frame(test_data)) {
    stop("test_data must be a data frame")
  }


  # Get column parameters from package environment
  id_column <- get("id_column", envir = .package_env)
  text_column <- get("text_column", envir = .package_env)

  # Validate that the specified columns exist
  if (id_column > ncol(test_data)) {
    stop("id_column (", id_column, ") is greater than the number of columns in test_data (", ncol(test_data), ")")
  }

  if (text_column > ncol(test_data)) {
    stop("text_column (", text_column, ") is greater than the number of columns in test_data (", ncol(test_data), ")")
  }

  # Get column names for easier reference
  id_col_name <- names(test_data)[id_column]
  text_col_name <- names(test_data)[text_column]

  # Validate that the text column contains character data
  if (!is.character(test_data[[text_column]]) && !is.factor(test_data[[text_column]])) {
    stop("Text column (column ", text_column, ") must contain character or factor data, but contains: ", class(test_data[[text_column]])[1])
  }

  # Check if text column has any non-NA values
  if (all(is.na(test_data[[text_column]]))) {
    stop("Text column (column ", text_column, ") contains only NA values. Please check your data.")
  }

  if (!is.numeric(slicing_n) || slicing_n <= 0) {
    stop("slicing_n must be a positive number")
  }

  if (!is.numeric(n_post) || n_post <= 0 || n_post > 1000) {
    stop("n_post must be a positive number between 1 and 1000")
  }

  if (!is.numeric(cores) || cores <= 0) {
    stop("cores must be a positive number")
  }

  # Check if cores is reasonable (not more than available cores + 1)
  available_cores <- parallel::detectCores()
  if (cores > available_cores + 1) {
    warning("Requested cores (", cores, ") exceeds available cores (", available_cores, "). This may cause performance issues.")
  }

  # Get provider from set_api_specs
  if (!exists("api_provider", envir = .package_env)) {
    stop("API provider not set. Please use set_api_specs() to configure your API settings.")
  }

  provider <- get("api_provider", envir = .package_env)
  if (!provider %in% c("OpenAI", "Groq")) {
    stop("Invalid provider. Provider must be 'OpenAI' or 'Groq' as set in set_api_specs()")
  }

  # Get API specifications to pass to workers
  api_model <- get("api_model", envir = .package_env)
  api_temp <- get("api_temp", envir = .package_env)
  multi_response <- get("multi_response", envir = .package_env)

  # Validate seed parameter
  if (!is.null(seed) && (!is.numeric(seed) || length(seed) != 1 || seed != as.integer(seed))) {
    stop("seed must be NULL or a single integer value")
  }

  # Calculate how many batches will have the larger size
  # For example: 967 posts, 8 cores -> 7 batches of 121, 1 batch of 120
  total_posts <- slicing_n
  base_batch_size <- floor(total_posts / cores)
  extra_posts <- total_posts %% cores

  if (extra_posts > 0) {
    if (verbose) {
      message("- First ", extra_posts, " batches will have ", base_batch_size + 1, " posts each")
      message("- Remaining ", cores - extra_posts, " batches will have ", base_batch_size, " posts each")
    }
  } else {
    if (verbose) message("All ", cores, " batches will have ", base_batch_size, " posts each")
  }

  #Making sure slicing_n is not greater than the number of rows in test_data
  if (slicing_n > nrow(test_data)) {
    stop("Make sure that slicing_n is not greater than the number of rows in test_data")
  }

  if (provider == "OpenAI") {
    test_answer <- ask_openai("What is the capital of Norway?")
    if (!is.na(test_answer)) {
      if (verbose) message("There are credits in the account!")
    } else {
      stop("OpenAI API test failed. Please check your API key and internet connection.")
    }
  } else if (provider == "Groq") {
    test_answer <- ask_groq("What is the capital of Norway?", temp = api_temp,
                            model = api_model, api_key = get("GROQ_API_KEY", envir = .package_env))
    if (!is.na(test_answer)) {
      if (verbose) message("There are credits in the account!")
    } else {
      stop("GROQ API test failed. Please check your API key and internet connection.")
    }
  }

  # Set seed for reproducible sampling if specified
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # First, add unique internal ID column to the full dataset
  test_data$internal_id <- 1:nrow(test_data)
  test_data <- test_data[, c("internal_id", setdiff(names(test_data), "internal_id"))]

  # Adjust column positions (add 1 because we added internal_id column)
  internal_id_column <- id_column + 1
  internal_text_column <- text_column + 1
  internal_id_col_name <- names(test_data)[internal_id_column]
  internal_text_col_name <- names(test_data)[internal_text_column]

  # Sample rows for processing
  subset_trial <- test_data %>% slice_sample(n = slicing_n, replace = FALSE)

  # Filter out posts with 4 or fewer characters (before counting missing IDs)
  original_count <- nrow(subset_trial)
  filtered_data <- subset_trial %>%
    filter(nchar(!!sym(internal_text_col_name)) > 4)
  filtered_count <- original_count - nrow(filtered_data)

  if (verbose) message("Filtered out ", filtered_count, " posts with 4 or fewer characters")

  # Check if we have any data left after filtering
  if (nrow(filtered_data) == 0) {
    stop("No data remaining after filtering posts with 4 or fewer characters. Please check your text column data.")
  }

  # Store all IDs being processed for later comparison using internal ID column
  all_processed_ids <- subset_trial$internal_id

  # Create batches with calculated sizes
  # First create a vector of batch sizes
  batch_sizes <- rep(base_batch_size, cores)
  if (extra_posts > 0) {
    batch_sizes[1:extra_posts] <- base_batch_size + 1
  }

  # Create cumulative indices for splitting
  batch_indices <- cumsum(c(0, batch_sizes))

  # Split the data into batches
  df_list <- list()
  for (i in 1:cores) {
    start_idx <- batch_indices[i] + 1
    end_idx <- batch_indices[i + 1]
    if (start_idx <= nrow(subset_trial)) {
      df_list[[i]] <- subset_trial[start_idx:min(end_idx, nrow(subset_trial)), , drop = FALSE]
    }
  }

  # Remove any empty batches
  df_list <- df_list[sapply(df_list, nrow) > 0]

  if(multi_core == TRUE) {
    plan(multicore, workers = cores)
  } else {
    plan(multisession, workers = cores) #Starts a multisession, using cores amount of cores
  }
  # Check if coding instruction and parameters are set in the main session
  if(!exists("coding_instruction", envir = .package_env) ||
            is.na(get("n_variables", envir = .package_env))) {
    stop("Coding instruction has not been set. Use set_coding_instruction() before running parallel_execute()")
  }

  if(!exists("n_variables", envir = .package_env) ||
      is.na(get("n_variables", envir = .package_env))) {
    stop("Variable number has not been set. Use set_parameters() before running parallel_execute()")
  }

  if(!exists("text_column", envir = .package_env) ||
      is.na(get("text_column", envir = .package_env))) {
    stop("Text column has not been set. Use set_parameters() before running parallel_execute()")
  }

  # Get the values from the main session's package environment
  coding_instruction <- get("coding_instruction", envir = .package_env)
  n_variables <- get("n_variables", envir = .package_env)
  id_column <- get("id_column", envir = .package_env)
  text_column <- get("text_column", envir = .package_env)
  sep <- get("sep", envir = .package_env)

  test_start <- Sys.time()

  # Ensure needed package is available for parallel helpers
  if (!requireNamespace("future.apply", quietly = TRUE)) {
    stop("Package 'future.apply' is required but not installed.")
  }

  # Check for other required packages
  required_packages <- c("dplyr", "stringr", "rlang")
  missing_packages <- required_packages[!sapply(required_packages, requireNamespace, quietly = TRUE)]
  if (length(missing_packages) > 0) {
    stop("Required packages are missing: ", paste(missing_packages, collapse = ", "), ". Please install them.")
  }

  # future.apply::future_lapply executes the loop body on the workers created
  # by the active future plan (multisession here). All referenced objects and
  # functions are automatically exported; no need to re-source files.

  result_list <- future_lapply(seq_along(df_list), function(i, track_progress, main_func, gpt_func,
                                                            groq_func, ask_openai, ask_groq) {

    # Use this worker's package environment (not the serialized one from the main session)
    worker_env <- get(".package_env", envir = asNamespace("rapidcodeR"))
    assign("coding_instruction", coding_instruction, envir = worker_env)
    assign("n_variables", n_variables, envir = worker_env)
    assign("id_column", id_column, envir = worker_env)
    assign("text_column", text_column, envir = worker_env)
    assign("sep", sep, envir = worker_env)
    assign("api_model", api_model, envir = worker_env)
    assign("api_temp", api_temp, envir = worker_env)
    assign("multi_response", multi_response, envir = worker_env)
    if (provider == "Groq" && exists("GROQ_API_KEY", envir = .package_env)) {
      assign("GROQ_API_KEY", get("GROQ_API_KEY", envir = .package_env), envir = worker_env)
    }

    # Pass the data and worker env explicitly to track_progress
    # Note: nrow(df_list[[i]]) is the batch size, n_post is the API call size
    track_progress(df_list[[i]], i, length(df_list), n_post, nrow(df_list[[i]]), provider, worker_env = worker_env, verbose = verbose)

  }, future.seed = TRUE, future.globals = list(
    `%>%` = `%>%`,
    filter = dplyr::filter,
    slice_sample = dplyr::slice_sample,
    arrange = dplyr::arrange,
    distinct = dplyr::distinct,
    bind_rows = dplyr::bind_rows,
    sym = rlang::sym,
    str_remove = stringr::str_remove,
    track_progress = track_progress,
    main_func = main_func,
    gpt_func = gpt_func,
    groq_func = groq_func,
    ask_openai = ask_openai,
    ask_groq = ask_groq,
    verbose = verbose
  ),
    track_progress = track_progress,
    main_func = main_func,
    gpt_func = gpt_func,
    groq_func = groq_func,
    ask_openai = ask_openai,
    ask_groq = ask_groq
  )

  plan(sequential) #Closes the multisession

  # Memory management: force garbage collection after parallel processing
  gc()

  # Process results from parallel execution
  processed_results <- lapply(result_list, function(x) {
    # Check if x is NULL or empty
    if (is.null(x) || length(x) == 0) {
      return(NULL)  # Return NULL for empty/failed results
    }

    # Handle nested list structure (x is a list of dataframes from main_func)
    if (is.list(x) && !is.data.frame(x)) {
      # Flatten the nested list and process each dataframe
      flattened_dataframes <- list()

      for (i in seq_along(x)) {
        if (!is.null(x[[i]]) && is.data.frame(x[[i]]) && nrow(x[[i]]) > 0) {
          # Process each dataframe in the nested list
          processed_df <- x[[i]]

          # Convert the first column to character (dynamically find it)
          if (ncol(processed_df) > 0) {
            first_col <- names(processed_df)[1]
            processed_df[[first_col]] <- as.character(processed_df[[first_col]])
          }

          # Convert the last column to character (dynamically find it)
          if (ncol(processed_df) > 0) {
            last_col_name <- names(processed_df)[ncol(processed_df)]
            processed_df[[last_col_name]] <- as.character(processed_df[[last_col_name]])
          }

          flattened_dataframes[[length(flattened_dataframes) + 1]] <- processed_df
        }
      }

      # Return the list of processed dataframes
      return(flattened_dataframes)
    }

    # Handle single dataframe (legacy case)
    if (is.data.frame(x) && nrow(x) > 0) {
      # Convert the first column to character (dynamically find it)
      if (ncol(x) > 0) {
        first_col <- names(x)[1]
        x[[first_col]] <- as.character(x[[first_col]])
      }

      # Convert the last column to character (dynamically find it)
      if (ncol(x) > 0) {
        last_col_name <- names(x)[ncol(x)]
        x[[last_col_name]] <- as.character(x[[last_col_name]])
      }

      return(list(x))  # Return as a list for consistency
    }

    return(NULL)  # Return NULL for any other case
  })

  # Filter out NULL results and bind
  valid_results <- processed_results[!sapply(processed_results, is.null)]

  # Check if we have any valid results
  failed_tasks <- length(result_list) - length(valid_results)
  if (length(valid_results) == 0) {
    warning("No valid results from parallel processing. All ", length(result_list), " tasks failed.")
    test_dfs <- data.frame()  # Create empty data frame
  } else {
    if (failed_tasks > 0) {
      warning(failed_tasks, " out of ", length(result_list), " tasks failed and were excluded from results.")
    }

    # Flatten all valid results into a single list of dataframes
    all_dataframes <- list()
    for (i in seq_along(valid_results)) {
      # Each element in valid_results is now a list of dataframes
      if (is.list(valid_results[[i]]) && length(valid_results[[i]]) > 0) {
        for (j in seq_along(valid_results[[i]])) {
          if (is.data.frame(valid_results[[i]][[j]]) && nrow(valid_results[[i]][[j]]) > 0) {
            all_dataframes[[length(all_dataframes) + 1]] <- valid_results[[i]][[j]]
          }
        }
      }
    }

    # Bind all dataframes together
    if (length(all_dataframes) > 0) {
      test_dfs <- do.call(bind_rows, all_dataframes)

      # Deduplicate results by internal_id only when multi_response=FALSE.
      # When multi_response=TRUE, multiple rows per ID are intentional (one post, multiple extracted responses).
      if (nrow(test_dfs) > 0 && !multi_response) {
        first_col_temp <- names(test_dfs)[1]
        duplicates_before <- nrow(test_dfs)
        test_dfs <- test_dfs %>% distinct(.data[[first_col_temp]], .keep_all = TRUE)
        duplicates_removed <- duplicates_before - nrow(test_dfs)
        if (duplicates_removed > 0 && verbose) {
          message("Removed ", duplicates_removed, " duplicate rows from initial results")
        }
      }
    } else {
      test_dfs <- data.frame()
    }
  }

  if (verbose) message("Initial results have ", nrow(test_dfs), " rows")

  # Compare processed results with original IDs to find missing ones
  if (nrow(test_dfs) > 0) {
    # Get the first column (internal_id) which contains the IDs
    first_col <- names(test_dfs)[1]
    # Deduplicate processed_ids to ensure accurate missing ID calculation
    # This prevents duplicate IDs from affecting the count of processed vs missing IDs
    processed_ids <- unique(as.numeric(test_dfs[[first_col]]))
    missing_ids <- all_processed_ids[!all_processed_ids %in% processed_ids]
    if (verbose) {
      message("Unique processed IDs: ", length(processed_ids))
      message("Missing IDs after parallel processing: ", length(missing_ids))
    }
    # Only call missingness_func if there are missing IDs
    if (length(missing_ids) > 0) {
      if (verbose) message("Attempting to recover ", length(missing_ids), " missing IDs")
      restored_dfs <- missingness_func(subset_trial, missing_ids, n_post, cores,
                                       provider, multi_core, verbose = verbose)
    } else {
      restored_dfs <- NULL
      if (verbose) message("No missing IDs to recover")
    }
  } else {
    # If no results at all, all IDs are missing
    missing_ids <- all_processed_ids
    if (verbose) message("No results from parallel processing. Attempting to recover all ", length(missing_ids), " IDs")
    restored_dfs <- missingness_func(subset_trial, missing_ids, n_post, cores, provider, multi_core, verbose = verbose)
  }

  process_time <- round(as.numeric(difftime(Sys.time(),  test_start, units = "secs")), 2)

  if (verbose) message("The processing took ", process_time, " seconds!")

  # Memory management: force garbage collection after recovery processing
  gc()

  # Combine results if recovery was successful
  if (!is.null(restored_dfs) && nrow(restored_dfs) > 0) {
    # Convert first column (internal_id) to character for binding
    if (nrow(test_dfs) > 0) {
      first_col <- names(test_dfs)[1]
      test_dfs[[first_col]] <- as.character(test_dfs[[first_col]])
    }
    first_col_restored <- names(restored_dfs)[1]
    restored_dfs[[first_col_restored]] <- as.character(restored_dfs[[first_col_restored]])

    all_data <- bind_rows(test_dfs, restored_dfs)
    if (verbose) message("Recovered ", nrow(restored_dfs), " additional rows")
    # Deduplicate after combining restored results only when multi_response=FALSE
    # (when multi_response=TRUE, multiple rows per ID are intentional)
    if (nrow(all_data) > 0 && !multi_response) {
      first_col_all <- names(all_data)[1]
      duplicates_before_combine <- nrow(all_data)
      all_data <- all_data %>% distinct(.data[[first_col_all]], .keep_all = TRUE)
      duplicates_removed_combine <- duplicates_before_combine - nrow(all_data)
      if (duplicates_removed_combine > 0 && verbose) {
        message("Removed ", duplicates_removed_combine, " duplicate rows after combining restored results")
      }
    }
  } else {
    all_data <- test_dfs
    if (verbose) message("No additional rows recovered")
  }

  if (verbose) message("Before final processing: ", nrow(all_data), " rows")

  # Sort by internal_id (first column) before removing it
  if (ncol(all_data) > 0) {
    internal_id_col <- names(all_data)[1]
    if (!is.na(internal_id_col)) {
      all_data <- all_data %>%
        arrange(as.numeric(.data[[internal_id_col]]))
    }
  }

  # Remove the internal_id column (first column) from the final output
  if (ncol(all_data) > 1) {
    output <- all_data[, -1, drop = FALSE]
  } else {
    output <- all_data
  }

  # Remove internal columns (multiple_response_flag, user_id) if present (no longer needed)
  internal_cols <- c("multiple_response_flag", "user_id")
  cols_to_drop <- internal_cols[internal_cols %in% names(output)]
  if (length(cols_to_drop) > 0) {
    output <- output[, setdiff(names(output), cols_to_drop)]
  }

  if (verbose) {
    message("After processing: ", nrow(output), " rows")
    message("Total posts with 4 or fewer characters filtered out: ", filtered_count)
  }

  if(benchmarking == FALSE) {
    return(output)
  } else {
    return(process_time)
  }

}

#' Process Missing Data with Parallel Execution
#'
#' This function handles data that failed to process in the initial parallel run
#' by reprocessing missing IDs with optimized batch sizes and parallel execution.
#'
#' @import future
#' @import future.apply
#' @importFrom dplyr filter slice_sample arrange distinct bind_rows
#' @importFrom dplyr "%>%" .data
#' @importFrom rlang sym
#'
#' @param subset_trial Data frame. The original data subset containing all posts.
#' @param missing_ids Numeric vector. IDs that failed processing and need retry.
#' @param n_post Integer. Number of posts to process in each API call.
#' @param cores Integer. Number of CPU cores available for parallel processing.
#' @param provider Character. AI provider to use, either "OpenAI" or "Groq".
#' @param multi_core Logical. If TRUE, uses multicore backend for parallel processing. If FALSE, uses multisession backend.
#' @param verbose Logical. If TRUE (default), progress and status messages are shown via \code{message()}. If FALSE, suppressed.
#'
#' @details
#' It continues until the number of missing items is reduced to a manageable level.
#' The function implements an adaptive retry strategy:
#' \itemize{
#'   \item Continues processing while more than 15 rows remain to be processed
#'   \item Filters out posts with 4 or fewer characters to avoid processing very short content
#'   \item Dynamically adjusts the number of parallel workers based on data size
#'   \item Ensures each worker has at least 15 posts to maintain efficiency
#'   \item Uses the same processing pipeline as the main parallel execution
#'   \item Accumulates successfully processed results across iterations
#'   \item Provides progress feedback on remaining items and filtered posts
#' }
#'
#' The function uses future-based parallel processing with automatic cleanup
#' and includes timing information for performance monitoring.
#'
#' @return Data frame. Successfully processed results from the missing data,
#'   or NULL if no missing data was recovered. The function also outputs
#'   debugging information about filtered posts with 4 or fewer characters.
#'
#' @seealso [parallel_execute()], [track_progress()]
#' @keywords internal
missingness_func <- function(subset_trial, missing_ids, n_post, cores, provider, multi_core = TRUE, verbose = TRUE) {

  # Start timing
  test_start <- Sys.time()

  # Input validation
  if (!is.data.frame(subset_trial)) {
    stop("subset_trial must be a data frame")
  }

  # Get column parameters from package environment
  text_column <- get("text_column", envir = .package_env)

  # Validate that the specified column exists
  if (text_column > ncol(subset_trial)) {
    stop("text_column (", text_column, ") is greater than the number of columns in subset_trial (", ncol(subset_trial), ")")
  }

  # Get column names for easier reference
  # ID column is always column 1 (internal_id)
  id_col_name <- names(subset_trial)[1]
  # Text column is adjusted by +1 because ID column was added
  internal_text_column <- text_column + 1
  text_col_name <- names(subset_trial)[internal_text_column]

  if (!is.numeric(missing_ids) || length(missing_ids) == 0) {
    stop("missing_ids must be a non-empty numeric vector")
  }

  if (!is.numeric(cores) || cores <= 0) {
    stop("cores must be a positive number")
  }

  if (!provider %in% c("OpenAI", "Groq")) {
    stop("provider must be either 'OpenAI' or 'Groq'")
  }

  # Define the %!in% operator for filtering
  `%!in%` <- Negate(`%in%`)

  # Get the values from the main session's package environment
  coding_instruction <- get("coding_instruction", envir = .package_env)
  n_variables <- get("n_variables", envir = .package_env)
  id_column <- get("id_column", envir = .package_env)
  text_column <- get("text_column", envir = .package_env)
  sep <- get("sep", envir = .package_env)

  # Get API specifications to pass to workers
  api_provider <- get("api_provider", envir = .package_env)
  api_model <- get("api_model", envir = .package_env)
  api_temp <- get("api_temp", envir = .package_env)
  multi_response <- get("multi_response", envir = .package_env)

  if (verbose) message("Starting recovery for ", length(missing_ids), " missing IDs")

  missing_data <- subset_trial %>%
    filter(!!sym(id_col_name) %in% missing_ids) %>%
    distinct(!!sym(id_col_name), .keep_all = TRUE)

  # Filter out posts with 4 or fewer characters using dynamic column name
  original_count <- nrow(missing_data)
  missing_data <- missing_data %>%
    filter(nchar(!!sym(text_col_name)) > 4)
  filtered_count <- original_count - nrow(missing_data)
  if (verbose) message("Filtered out ", filtered_count, " posts with 4 or fewer characters in recovery")

  if (verbose) message("Found ", nrow(missing_data), " rows to recover")

  # If no data found, return NULL
  if (nrow(missing_data) == 0) {
    if (verbose) message("No data found for missing IDs after filtering")
    return(NULL)
  }

  # Track total filtered posts across iterations
  total_filtered <- filtered_count

  # Run recovery iterations until fewer than 15 rows remain
  while(nrow(missing_data) > 15) {

    if (verbose) message("Processing recovery batch of ", nrow(missing_data), " rows")

    rep_end <- cores

    while(nrow(missing_data) / rep_end <= 15 && rep_end > 1) {
      rep_end <- rep_end - 1
    }

    if(multi_core == TRUE) {
      plan(multicore, workers = rep_end)
    } else {
      plan(multisession, workers = rep_end)
    }

    # Ensure future.apply is available
    if (!requireNamespace("future.apply", quietly = TRUE)) {
      stop("Package 'future.apply' is required but not installed.")
    }

    missing_list <- split(missing_data, rep(1:rep_end, length.out = nrow(missing_data)))

    result_list <- future_lapply(seq_along(missing_list), function(i, track_progress, main_func, gpt_func,
                                                                  groq_func, ask_openai, ask_groq) {

      # Use this worker's package environment (not the serialized one from the main session)
      worker_env <- get(".package_env", envir = asNamespace("rapidcodeR"))
      assign("coding_instruction", coding_instruction, envir = worker_env)
      assign("n_variables", n_variables, envir = worker_env)
      assign("id_column", id_column, envir = worker_env)
      assign("text_column", text_column, envir = worker_env)
      assign("sep", sep, envir = worker_env)
      assign("api_provider", api_provider, envir = worker_env)
      assign("api_model", api_model, envir = worker_env)
      assign("api_temp", api_temp, envir = worker_env)
      assign("multi_response", multi_response, envir = worker_env)
      if (provider == "Groq" && exists("GROQ_API_KEY", envir = .package_env)) {
        assign("GROQ_API_KEY", get("GROQ_API_KEY", envir = .package_env), envir = worker_env)
      }

      # Pass worker env and verbose so track_progress can gate messages
      track_progress(missing_list[[i]], i, length(missing_list), n_post, nrow(missing_list[[i]]), provider, worker_env = worker_env, verbose = verbose)
    }, future.seed = TRUE, future.globals = list(
      `%>%` = `%>%`,
      filter = dplyr::filter,
      slice_sample = dplyr::slice_sample,
      arrange = dplyr::arrange,
      distinct = dplyr::distinct,
      bind_rows = dplyr::bind_rows,
      sym = rlang::sym,
      str_remove = stringr::str_remove,
      track_progress = track_progress,
      main_func = main_func,
      gpt_func = gpt_func,
      groq_func = groq_func,
      ask_openai = ask_openai,
      ask_groq = ask_groq,
      verbose = verbose
    ),
      track_progress = track_progress,
      main_func = main_func,
      gpt_func = gpt_func,
      groq_func = groq_func,
      ask_openai = ask_openai,
      ask_groq = ask_groq
    )

    if (verbose) message("The recovery took ", round(as.numeric(difftime(Sys.time(),  test_start, units = "secs")), 2), " seconds!")

    plan(sequential)

    # Memory management: force garbage collection after parallel processing
    gc()

    # Process results and filter out NULL/empty results
    processed_results <- lapply(result_list, function(x) {
      # Check if x is NULL or empty
      if (is.null(x) || length(x) == 0) {
        return(NULL)  # Return NULL for empty/failed results
      }

      # Handle nested list structure (x is a list of dataframes from main_func)
      if (is.list(x) && !is.data.frame(x)) {
        # Flatten the nested list and process each dataframe
        flattened_dataframes <- list()

        for (i in seq_along(x)) {
          if (!is.null(x[[i]]) && is.data.frame(x[[i]]) && nrow(x[[i]]) > 0) {
            # Process each dataframe in the nested list
            processed_df <- x[[i]]

            # Convert the first column to character (dynamically find it)
            if (ncol(processed_df) > 0) {
              first_col <- names(processed_df)[1]
              processed_df[[first_col]] <- as.character(processed_df[[first_col]])
            }

            # Convert the last column to character (dynamically find it)
            if (ncol(processed_df) > 0) {
              last_col_name <- names(processed_df)[ncol(processed_df)]
              processed_df[[last_col_name]] <- as.character(processed_df[[last_col_name]])
            }

            flattened_dataframes[[length(flattened_dataframes) + 1]] <- processed_df
          }
        }

        # Return the list of processed dataframes
        return(flattened_dataframes)
      }

      # Handle single dataframe (legacy case)
      if (is.data.frame(x) && nrow(x) > 0) {
        # Convert the first column to character (dynamically find it)
        if (ncol(x) > 0) {
          first_col <- names(x)[1]
          x[[first_col]] <- as.character(x[[first_col]])
        }

        # Convert the last column to character (dynamically find it)
        if (ncol(x) > 0) {
          last_col_name <- names(x)[ncol(x)]
          x[[last_col_name]] <- as.character(x[[last_col_name]])
        }

        return(list(x))  # Return as a list for consistency
      }

      return(NULL)  # Return NULL for any other case
    })

    # Filter out NULL results and bind
    valid_results <- processed_results[!sapply(processed_results, is.null)]

    failed_tasks <- length(result_list) - length(valid_results)
    if (length(valid_results) == 0) {
      temp_df <- data.frame()  # Create empty data frame
      if (verbose) message("No valid results from recovery iteration")
    } else {
      if (failed_tasks > 0 && verbose) {
        message(failed_tasks, " out of ", length(result_list), " missingness recovery tasks failed.")
      }

      # Handle list of dataframes from each task
      all_dataframes <- list()
      for (i in seq_along(valid_results)) {
        # Each element in valid_results is now a list of dataframes
        if (is.list(valid_results[[i]]) && length(valid_results[[i]]) > 0) {
          for (j in seq_along(valid_results[[i]])) {
            if (is.data.frame(valid_results[[i]][[j]]) && nrow(valid_results[[i]][[j]]) > 0) {
              all_dataframes[[length(all_dataframes) + 1]] <- valid_results[[i]][[j]]
            }
          }
        }
      }

      # Bind all dataframes together
      if (length(all_dataframes) > 0) {
        temp_df <- do.call(bind_rows, all_dataframes)
      } else {
        temp_df <- data.frame()
      }

      if (verbose) message("Recovered ", nrow(temp_df), " rows in this iteration")
    }

    if(exists("restored_dfs")) {
      if (nrow(temp_df) > 0) {
        restored_dfs <- bind_rows(restored_dfs, temp_df)
      }
      # If temp_df is empty, just keep restored_dfs as is
    } else {
      restored_dfs <- temp_df
    }

    if (verbose) message("Recovery iteration completed")

    # Update missing_data for next iteration by removing successfully processed IDs
    if (nrow(temp_df) > 0) {
      # Get the first column (internal_id) which contains the IDs
      first_col <- names(temp_df)[1]
      processed_ids <- as.numeric(temp_df[[first_col]])

      missing_data <- missing_data %>%
        filter(!!sym(id_col_name) %!in% processed_ids)

      # Filter out posts with 4 or fewer characters for next iteration using dynamic column name
      original_count <- nrow(missing_data)
      missing_data <- missing_data %>%
        filter(nchar(!!sym(text_col_name)) > 4)
      iteration_filtered <- original_count - nrow(missing_data)
      total_filtered <- total_filtered + iteration_filtered

      if (iteration_filtered > 0 && verbose) {
        message("Filtered out ", iteration_filtered, " additional posts with 4 or fewer characters")
      }

      if (verbose) message("Remaining rows for next iteration: ", nrow(missing_data))
    } else {
      if (verbose) message("No results from this iteration, stopping recovery")
      break
    }
  }

  # Check if we stopped because of the threshold
  if (nrow(missing_data) <= 15 && nrow(missing_data) > 0 && verbose) {
    message("Recovery stopped because only ", nrow(missing_data), " rows remain (below threshold of 15)")
  }



  if(exists("restored_dfs") && nrow(restored_dfs) > 0){
    if (verbose) {
      message("Recovery completed. Returning ", nrow(restored_dfs), " recovered rows")
      message("Total posts with 4 or fewer characters filtered out in recovery: ", total_filtered)
    }
    return(restored_dfs)
  } else {
    if (verbose) {
      message("Recovery completed. No rows recovered.")
      message("Total posts with 4 or fewer characters filtered out in recovery: ", total_filtered)
    }
    return(NULL)
  }

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groq_function.R
\name{groq_func}
\alias{groq_func}
\title{Process Text Data Using GROQ AI API}
\usage{
groq_func(data_subset, n_post, coding_instruction, worker_env = NULL)
}
\arguments{
\item{data_subset}{Data frame. Subset of text data to process. Must contain
columns specified by \code{id_column} and \code{text_column} parameters (set via \code{set_parameters()}).}

\item{n_post}{Integer. Number of posts to process in this batch.}

\item{coding_instruction}{Character. Instructions for the AI model specifying
how to extract and format variables from the text.}

\item{worker_env}{Environment or NULL. When running in a parallel worker, the worker's
package environment. If NULL (default), \code{.package_env} is used.}
}
\value{
Data frame. Each row contains the AI's response for one post,
with columns named Var1, Var2, etc. Returns NULL if no posts remain after filtering.
}
\description{
This function sends text data to the GROQ AI API for variable extraction,
optimized for speed and cost efficiency. It formats the data for API consumption
and parses the CSV-formatted response.
}
\details{
The function performs several optimizations:
\itemize{
\item Filters out very short posts (< 4 characters) to reduce noise
\item Sanitizes text by replacing problematic characters (semicolons, quotes)
\item Constructs a single prompt with multiple posts for batch processing
\item Calls GROQ API with parameters from set_api_specs() (e.g. temperature, top_p)
\item Parses the response assuming CSV format with newline separators
\item Always returns a dataframe (single rows become 1-row dataframes)
}

The function expects the AI to return one line per input post in CSV format.
Trailing semicolons are automatically removed from the response.
}
\seealso{
\code{\link[=ask_groq]{ask_groq()}}, \code{\link[=gpt_func]{gpt_func()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_progress.R
\name{track_progress}
\alias{track_progress}
\title{Track Progress of Parallel Processing Tasks}
\usage{
track_progress(
  df,
  index,
  total_tasks,
  n_post,
  batch_size,
  provider,
  worker_env = NULL,
  verbose = TRUE
)
}
\arguments{
\item{df}{Data frame. A subset of data to process in this task.}

\item{index}{Integer. The current task number (for progress tracking).}

\item{total_tasks}{Integer. Total number of tasks in the parallel job.}

\item{n_post}{Integer. Number of posts to process in each API call.}

\item{provider}{Character. AI provider to use, either "OpenAI" or "Groq".}

\item{worker_env}{Environment or NULL. When running in a parallel worker, the worker's
package environment so config (e.g. \code{coding_instruction}, \code{id_column})
is available. If NULL (default), the package's \code{.package_env} is used.}

\item{verbose}{Logical. If TRUE (default), progress messages are shown. If FALSE, suppressed.}
}
\value{
Data frame with successfully processed results from \code{main_func()},
or NULL if an error occurs during processing.
}
\description{
This function wraps the main processing function with progress tracking,
error handling, and rate limiting. It's designed to be called from parallel
workers to process data batches while providing feedback on progress.
}
\details{
The function provides several important features:
\itemize{
\item Progress messages showing current task vs total tasks
\item Rate limiting with 5-second delays to avoid overwhelming APIs
\item Comprehensive error handling with informative error messages
\item Automatic validation of provider parameter
\item Graceful degradation returning NULL on errors
}

This function is primarily used internally by the parallel processing
framework and calls \code{main_func()} with the \code{n_post} parameter passed by the caller.
}
\seealso{
\code{\link[=main_func]{main_func()}}, \code{\link[=parallel_execute]{parallel_execute()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{set_api_specs}
\alias{set_api_specs}
\title{Set API Specifications for AI Models}
\usage{
set_api_specs(provider, model, temp, api_key, multi_response = FALSE)
}
\arguments{
\item{provider}{Character. The AI provider to use, either "OpenAI" or "Groq".}

\item{model}{Character. The specific model to use (e.g., "gpt-4", "gpt-3.5-turbo", "llama-3.3-70b-versatile").}

\item{temp}{Numeric. Temperature parameter controlling randomness (0-1). Lower values make output more deterministic.}

\item{api_key}{Character. The API key for the specified provider.}

\item{multi_response}{Logical. If TRUE, allows AI to return multiple rows per input (only works with n_post=1).
If FALSE (default), AI must return exactly one row per input.}
}
\value{
Character. A confirmation message.
}
\description{
This function sets the API specifications including provider, model, temperature,
and API key for AI model interactions. It replaces the need for separate
set_api_keys calls and allows for more flexible configuration.
}
\details{
This function stores the API specifications in the package's internal environment
and sets the appropriate environment variables for API authentication. The
specifications are automatically retrieved by processing functions.
}
\examples{
\dontrun{
# Set OpenAI specifications
set_api_specs(
  provider = "OpenAI",
  model = "gpt-4",
  temp = 0.7,
  api_key = "your_openai_key"
)

# Set GROQ specifications
set_api_specs(
  provider = "Groq",
  model = "llama-3.3-70b-versatile",
  temp = 0.3,
  api_key = "your_groq_key"
)
}

}
\seealso{
\code{\link[=set_coding_instruction]{set_coding_instruction()}}, \code{\link[=set_parameters]{set_parameters()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/help_functions.R
\name{make_value_row}
\alias{make_value_row}
\title{Parse AI Response into Formatted Row String}
\usage{
make_value_row(ai_response)
}
\arguments{
\item{ai_response}{Character. A single response string from an AI model, typically
containing semicolon-separated values.}
}
\value{
Character. A formatted row string in SQL format like "('val1','val2',...)",
or NA if validation fails.
}
\description{
This function takes a single AI response string and formats it into a standardized
SQL-style row format for consistent data processing. It handles text sanitization
and ensures the response has exactly N elements (N set by set_parameters()).
}
\details{
The function performs several formatting operations:
\itemize{
\item Converts input to character format
\item Replaces quotes with backticks to avoid SQL conflicts
\item Splits on semicolons to extract individual values
\item Trims whitespace from each value
\item Pads with NA values if fewer than expected elements
\item Validates that we have the expected number of elements
\item Converts to lowercase and formats as SQL-style row
}

The function expects exactly N values as set by set_parameters().
}
\examples{
\dontrun{
# AI response with semicolon-separated values (assuming set_parameters(n_variables = 6))
ai_response <- "123;en;Hello world;positive;high;topic1"
formatted_row <- make_value_row(ai_response)
# Returns: "('123','en','hello world','positive','high','topic1')"
}

}
\seealso{
\code{\link[=main_func]{main_func()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_overlap.R
\name{calculate_overlap}
\alias{calculate_overlap}
\title{Calculate Inter-Rater Agreement Across Multiple Datasets}
\usage{
calculate_overlap(datasets, alpha = FALSE, key_cols = NULL)
}
\arguments{
\item{datasets}{List. A list of data frames to compare. Each data frame should
have the same column structure with variable columns (Var1, Var2, etc.).}

\item{alpha}{Logical. If TRUE, also compute Krippendorff's Alpha (nominal)
for each variable across the provided datasets. Defaults to FALSE.}

\item{key_cols}{Character vector or NULL. Column name(s) used to match rows
across datasets. Use this when datasets have different row counts due to a
compound row identity (e.g. \code{key_cols = c("Var1", "Var3")}). Rows with
NA in any key column are dropped before matching. When NULL (default), rows
are matched on the first column only, which is the standard behaviour for
single-response output from \code{parallel_execute()}.}
}
\value{
Matrix. If alpha = FALSE: one-row matrix with average percentage
agreement for each variable (columns), rounded to 2 decimals.
If alpha = TRUE: two-row matrix with row "Overlap" (percent agreement)
and row "Alpha" (Krippendorff's Alpha for nominal data), rounded
to 3 decimals for Alpha.
}
\description{
This function calculates the percentage overlap (agreement) between multiple
datasets containing the same variables. It's designed to assess reliability
and consistency when the same data is processed multiple times or by different
systems/raters.
}
\details{
The function performs comprehensive overlap analysis:
\itemize{
\item Finds common rows across all datasets based on \code{key_cols}
\item Aligns and sorts datasets by the key before comparing
\item Calculates pairwise agreement for all possible dataset combinations
\item Computes variable-wise agreement percentages
\item Returns average agreement across all dataset pairs
}

Agreement is calculated as the percentage of cases where values match exactly.
The number of rows used in the comparison is printed as a message (\code{N = ...}).
}
\examples{
\dontrun{
# Standard use: compare two processing runs (single-response output)
run1 <- parallel_execute(my_data, slicing_n = 100, cores = 4)
run2 <- parallel_execute(my_data, slicing_n = 100, cores = 4)
calculate_overlap(list(run1, run2))

# Multi-response output: datasets may differ in row count;
# match on compound key (e.g. speech ID + paragraph number)
calculate_overlap(list(run1, run2), key_cols = c("Var1", "Var3"))
}

}
\seealso{
\code{\link[=parallel_execute]{parallel_execute()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel_execute.R
\name{parallel_execute}
\alias{parallel_execute}
\title{Execute Parallel Text Processing with AI Models}
\usage{
parallel_execute(
  test_data,
  slicing_n,
  n_post = 15,
  cores = 8,
  seed = NULL,
  multi_core = FALSE,
  benchmarking = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{test_data}{Data frame. Input data containing text to be processed.}

\item{slicing_n}{Integer. Number of rows to sample from the input data for processing.
The batch size will be automatically calculated by dividing this by the number of cores.}

\item{n_post}{Integer. Number of posts to process in each API call. Default is 15.
Must be between 1 and 1000. Larger values may be more efficient but use more API credits.}

\item{cores}{Integer. Number of CPU cores to use for parallel processing. Default is 8.
The function will use the provider specified in set_api_specs() to determine
whether to use OpenAI or GROQ services.}

\item{seed}{Integer or NULL. Random seed for reproducible sampling. If NULL (default),
sampling is random. If set to a number, ensures identical datasets across runs.}

\item{multi_core}{Logical. If TRUE, uses multicore backend for parallel processing
(faster but Unix/Mac only). If FALSE (default), uses multisession backend
(works on all platforms including Windows).}

\item{benchmarking}{Logical. If TRUE, returns processing time in seconds instead of
the result data frame. Default is FALSE.}

\item{verbose}{Logical. If TRUE (default), progress and status messages are
printed via \code{message()}. If FALSE, such messages are suppressed.}
}
\value{
Data frame. Processed results containing extracted variables,
with duplicates removed (when \code{multi_response=FALSE}) and sorted by ID.
}
\description{
This is the main function that orchestrates parallel processing of text data
using AI APIs for variable extraction. It optimizes for speed and cost by
distributing work across multiple cores and handling missing data efficiently.
}
\details{
The function implements a multi-stage parallel processing workflow:
\enumerate{
\item Validates parameters and tests API connectivity
\item Randomly samples data and splits into batches for parallel processing
\item Executes parallel processing using futures and tracks progress
\item Handles missing/failed extractions with a secondary processing stage
\item Combines and deduplicates results
}

The function requires a coding instruction to be set via \code{set_coding_instruction()}
before execution. This instruction tells the AI how to format its responses.
}
\examples{
\dontrun{
# Set up API specifications and coding instruction
set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_key")
set_coding_instruction("Extract sentiment: id,sentiment,confidence")

# Process data with random sampling (default)
results <- parallel_execute(
  test_data = my_data,
  slicing_n = 240,
  cores = 4
)

# Process data with reproducible sampling (for calculate_overlap)
results1 <- parallel_execute(
  test_data = my_data,
  slicing_n = 240,
  cores = 4,
  seed = 123
)

# Switch to GROQ for comparison
set_api_specs(provider = "Groq", model = "llama-3.3-70b-versatile",
  temp = 0.3, api_key = "your_groq_key")
results2 <- parallel_execute(
  test_data = my_data,
  slicing_n = 240,
  cores = 4,
  seed = 123
)

# Calculate overlap between identical datasets
overlap_scores <- calculate_overlap(list(results1, results2))

}

}
\seealso{
\code{\link[=set_api_specs]{set_api_specs()}}, \code{\link[=set_coding_instruction]{set_coding_instruction()}}, \code{\link[=set_parameters]{set_parameters()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{package}
\name{rapidcodeR-package}
\alias{rapidcodeR}
\alias{rapidcodeR-package}
\title{rapidcodeR: Optimized Data Analysis System for AI-Based Text Processing}
\description{
A high-performance R package for extracting machine-readable variables from
natural language text using AI APIs. The package is optimized for speed and
cost efficiency through parallel processing and direct CSV-formatted responses
from language models.
}
\section{Funding}{

This work is financed in large part by the Research Council of Norway
(grant #324931).
}

\section{Key Features}{

\itemize{
\item \strong{Parallel Processing}: Distributes work across multiple CPU cores for maximum speed
\item \strong{Cost Optimization}: Uses batch processing to minimize API calls
\item \strong{Multi-Provider Support}: Works with both OpenAI GPT and GROQ AI services
\item \strong{Robust Error Handling}: Automatic retry mechanisms for failed extractions
\item \strong{CSV-Optimized Output}: AI models return structured data for efficient parsing
\item \strong{Missing Data Recovery}: Secondary processing stage for failed extractions
}
}

\section{Main Functions}{

\describe{
\item{\code{\link{parallel_execute}}}{Main function for parallel text processing}
\item{\code{\link{set_api_specs}}}{Configure API specifications for AI services}
\item{\code{\link{set_coding_instruction}}}{Define how AI should extract variables}
\item{\code{\link{set_parameters}}}{Configure processing parameters}
\item{\code{\link{calculate_overlap}}}{Assess reliability across multiple runs}
}
}

\section{Workflow}{

\enumerate{
\item Set up API specifications using \code{set_api_specs()}
\item Configure processing parameters with \code{set_parameters()} (optional)
\item Define extraction instructions with \code{set_coding_instruction()}
\item Process data with \code{parallel_execute()}
\item Optionally assess reliability with \code{calculate_overlap()}
}
}

\section{Data Format}{

Input data should contain columns specified by \code{set_parameters()}:
\itemize{
\item ID column: Unique identifier for each text (default column 1)
\item Text column: The text content to be processed (default column 2)
}

Output contains configurable columns: id, language, text, and N extracted variables (Var1-VarN).
}

\examples{
\dontrun{
# Basic workflow
library(rapidcodeR)

# 1. Set up API specifications
set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_openai_key")

# 2. Configure processing parameters (optional)
set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ",")

# 3. Define what to extract
instruction <- paste(
  "Extract sentiment and topic from each post.",
  "Return as CSV: id,language,text,sentiment,confidence,topic,relevance"
)
set_coding_instruction(instruction)

# 4. Process your data
results <- parallel_execute(
  test_data = my_text_data,
  slicing_n = 240,     # Process 240 random posts
  cores = 4,           # Use 4 CPU cores
  # Provider determined by set_api_specs()
)

# 5. Check reliability (optional)
run1 <- parallel_execute(my_data, 100, cores = 4, seed = 123)
run2 <- parallel_execute(my_data, 100, cores = 4, seed = 123)
agreement <- calculate_overlap(list(run1, run2))
}

}
\author{
\strong{Maintainer}: Gabriel Lönn \email{g.e.lonn@stv.uio.no}

Authors:
\itemize{
  \item Gabriel Lönn \email{g.e.lonn@stv.uio.no}
}

Other contributors:
\itemize{
  \item Sebastian Schutte (Original code and package idea provided by.) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/help_functions.R
\name{ask_openai}
\alias{ask_openai}
\title{Send Prompt to OpenAI API}
\usage{
ask_openai(prompt, temp = NULL, topp = 1, model = NULL)
}
\arguments{
\item{prompt}{Character. The text prompt to send to the AI model.}

\item{temp}{Numeric. Temperature parameter controlling randomness (0-1).
Lower values make output more deterministic. If NULL, retrieved from set_api_specs().}

\item{topp}{Numeric. Top-p parameter controlling diversity (0-1). Default is 1.}

\item{model}{Character. OpenAI model to use. If NULL, retrieved from set_api_specs().}
}
\value{
Character. The AI model's response to the prompt, or NA if an error occurs.
}
\description{
This function sends a text prompt to the OpenAI API and returns the model's
response. It includes robust error handling and uses the openai package for API communication.
}
\details{
The function uses the openai package to communicate with OpenAI's API. It includes
comprehensive error handling that returns NA if the API call fails or returns
an unexpected format. The function expects the OPENAI_API_KEY environment variable
to be set via \code{set_api_specs()}.
}
\examples{
\dontrun{
# Set API specifications first
set_api_specs(provider = "OpenAI", model = "gpt-4", temp = 0.7, api_key = "your_api_key")

# Send a simple prompt
response <- ask_openai("What is the capital of France?")

# Use different parameters
response <- ask_openai("Analyze this text", temp = 0.3, topp = 0.9)
}

}
\seealso{
\code{\link[=ask_groq]{ask_groq()}}, \code{\link[=set_api_specs]{set_api_specs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main_func.R
\name{main_func}
\alias{main_func}
\title{Main Text Processing Function for AI-Based Variable Extraction}
\usage{
main_func(df, to_code_max_id, n_post, provider, worker_env = NULL)
}
\arguments{
\item{df}{Data frame. Input data subset containing text to be processed.}

\item{to_code_max_id}{Integer. Maximum number of posts to process in this batch.}

\item{n_post}{Integer. Number of posts to process per API call (typically 15).}

\item{provider}{Character. AI provider to use, either "OpenAI" or "Groq".}

\item{worker_env}{Environment or NULL. When running in a parallel worker, the worker's
package environment. If NULL (default), \code{.package_env} is used.}
}
\value{
List of data frames. Each element contains successfully processed results
with extracted variables. Returns empty list if no successful processing.
}
\description{
This function processes batches of text data using AI models to extract
machine-readable variables. It implements robust error handling and retry
logic to ensure reliable processing even with API failures.
}
\details{
The function implements a robust processing loop that:
\itemize{
\item Samples posts randomly to avoid processing order bias
\item Makes API calls in manageable batches
\item Validates and cleans AI responses
\item Handles API failures with counter and early stopping
\item Tracks missing IDs for potential reprocessing
\item Collects dataframes into a list for later binding
}

The AI models are instructed to return data in CSV format for efficient parsing.
The function expects responses with exactly N columns as set by set_parameters().
}
\examples{
\dontrun{
# For set_parameters(n_variables = 6):
instruction <- "Extract variables: var1,var2,var3,var4,var5,var6"
result <- main_func(
  df = my_data_subset,
  to_code_max_id = 30,
  n_post = 15,
  provider = "OpenAI"
)
processed_data <- bind_rows(result)  # Combine all dataframes
}

}
\seealso{
\code{\link[=gpt_func]{gpt_func()}}, \code{\link[=groq_func]{groq_func()}}, \code{\link[=make_value_row]{make_value_row()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{set_coding_instruction}
\alias{set_coding_instruction}
\title{Set Coding Instructions for AI Models}
\usage{
set_coding_instruction(instruction)
}
\arguments{
\item{instruction}{Character. A string containing the coding instruction that
tells the AI model how to process and format the text data. Should specify
the expected output format (typically CSV-structured responses).}
}
\value{
Character. A confirmation message indicating the instruction has been set.
}
\description{
This function sets the coding instructions that will be used by AI models
to extract machine-readable variables from natural language text. The instruction
defines how the AI should format and structure its responses.
}
\details{
The coding instruction is stored in the package's internal environment and
is automatically retrieved by processing functions. This instruction typically
defines:
\itemize{
\item Expected output format (e.g., CSV structure)
\item Variable definitions and coding schemes
\item Response formatting requirements
}
}
\examples{
\dontrun{
instruction <- "Extract sentiment and topics from posts. Return as CSV: id,sentiment,topic"
set_coding_instruction(instruction)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{set_parameters}
\alias{set_parameters}
\title{Set Processing Parameters for AI Models}
\usage{
set_parameters(n_variables = 9, id_column = 1, text_column = 2, sep = ";")
}
\arguments{
\item{n_variables}{Integer. Number of columns to extract (default is 9).
This determines the total number of columns in the final dataset.
Must be between 1 and 20.}

\item{id_column}{Integer. The column number containing the unique ID for each
observation (default is 1). Must be a positive integer.}

\item{text_column}{Integer. The column number containing the text content
to be processed (default is 2). Must be a positive integer.}

\item{sep}{Character. Separator used in API response parsing (default is ";").
Must be one of: ";" (semicolon), "," (comma), or "|" (vertical bar).}
}
\value{
Character. A confirmation message indicating the parameters have been set.
}
\description{
This function configures the parameters that will be used by AI models
to extract machine-readable variables from natural language text.
}
\details{
The function sets the expected parameters in the package's internal
environment. This affects:
\itemize{
\item Response parsing and validation
\item Column naming (Var1, Var2, ..., VarN)
\item Data structure expectations across all processing functions
\item Column identification for text content and unique IDs
\item Separator used for parsing API responses
}

The package automatically adds a unique internal ID column to each row during processing.
This internal ID is used for tracking and is removed from the final output.

The total expected columns will be n_variables:
\itemize{
\item All columns: Var1, Var2, ..., VarN
}
}
\examples{
\dontrun{
# Extract 4 columns, ID in column 1, text in column 2, semicolon separator
set_parameters(n_variables = 4, id_column = 1, text_column = 2, sep = ";")

# Extract 10 columns, ID in column 3, text in column 5, comma separator
set_parameters(n_variables = 10, id_column = 3, text_column = 5, sep = ",")

# Extract 6 columns, ID in column 2, text in column 3, vertical bar separator
set_parameters(n_variables = 6, id_column = 2, text_column = 3, sep = "|")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingness_func.R
\name{missingness_func}
\alias{missingness_func}
\title{Process Missing Data with Parallel Execution}
\usage{
missingness_func(
  subset_trial,
  missing_ids,
  n_post,
  cores,
  provider,
  multi_core = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{subset_trial}{Data frame. The original data subset containing all posts.}

\item{missing_ids}{Numeric vector. IDs that failed processing and need retry.}

\item{n_post}{Integer. Number of posts to process in each API call.}

\item{cores}{Integer. Number of CPU cores available for parallel processing.}

\item{provider}{Character. AI provider to use, either "OpenAI" or "Groq".}

\item{multi_core}{Logical. If TRUE, uses multicore backend for parallel processing. If FALSE, uses multisession backend.}

\item{verbose}{Logical. If TRUE (default), progress and status messages are shown via \code{message()}. If FALSE, suppressed.}
}
\value{
Data frame. Successfully processed results from the missing data,
or NULL if no missing data was recovered. The function also outputs
debugging information about filtered posts with 4 or fewer characters.
}
\description{
This function handles data that failed to process in the initial parallel run
by reprocessing missing IDs with optimized batch sizes and parallel execution.
}
\details{
It continues until the number of missing items is reduced to a manageable level.
The function implements an adaptive retry strategy:
\itemize{
\item Continues processing while more than 15 rows remain to be processed
\item Filters out posts with 4 or fewer characters to avoid processing very short content
\item Dynamically adjusts the number of parallel workers based on data size
\item Ensures each worker has at least 15 posts to maintain efficiency
\item Uses the same processing pipeline as the main parallel execution
\item Accumulates successfully processed results across iterations
\item Provides progress feedback on remaining items and filtered posts
}

The function uses future-based parallel processing with automatic cleanup
and includes timing information for performance monitoring.
}
\seealso{
\code{\link[=parallel_execute]{parallel_execute()}}, \code{\link[=track_progress]{track_progress()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/help_functions.R
\name{ask_groq}
\alias{ask_groq}
\title{Send Prompt to GROQ AI API}
\usage{
ask_groq(prompt, temp, topp = 1, model, api_key)
}
\arguments{
\item{prompt}{Character. The text prompt to send to the AI model.}

\item{temp}{Numeric. Temperature parameter controlling randomness (0-1).
Lower values make output more deterministic. Required (typically from set_api_specs()).}

\item{topp}{Numeric. Top-p parameter controlling diversity (0-1). Default is 1.}

\item{model}{Character. GROQ model to use. Required (typically from set_api_specs()).}

\item{api_key}{Character. GROQ API key. Must be provided.}
}
\value{
Character. The AI model's response to the prompt, or NA if an error occurs.
}
\description{
This function sends a text prompt to the GROQ AI API and returns the model's
response. It includes error handling and uses the groqR package for API communication.
}
\details{
The function uses the groqR package to communicate with GROQ's API. It includes
error handling that returns NA if the API call fails. The function is optimized
for text processing tasks with a maximum token limit of 5000.
}
\examples{
\dontrun{
# Set API key first
Sys.setenv(GROQ_API_KEY = "your_api_key")

# Send a simple prompt
response <- ask_groq("What is the capital of France?")

# Use different parameters
response <- ask_groq("Analyze this text", temp = 0.3, topp = 0.9)
}

}
\seealso{
\code{\link[=ask_openai]{ask_openai()}}
}
